//  Copyright (c) 2005, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "shader.h"
#include "geoshader.h"

typedef struct
{
	miColor	color_A;
	miColor	color_B;
	miTag object_A;
	miTag object_B;
	miBoolean swap_AB;
	int mode;
	miBoolean check_camera;
}p_CSG_t;

typedef struct
{
	miBoolean inside_a;			//current A value
	miBoolean inside_b;			//current B value
	miBoolean p_inside_a;		//previous value of A, we can restore if need (mostly used for shadow shader)
	miBoolean p_inside_b;		//previous value of B, we can restore if need (mostly used for shadow shader)
}inside;

typedef struct
{
	miBoolean inside_a;
	miBoolean inside_b;
	miBoolean p_inside_a;
	miBoolean p_inside_b;
}inside_shd;

static void choose_volume(miState *state)
{
	register int		n = 0;		/* same-material counter */
	register miState	*s, *s_in = 0;	/* for finding enclosing mtl */

	for (n=0, s=state; s; s=s->parent)		/* history? */
		if ((s->type == miRAY_TRANSPARENT	   ||
		     s->type == miRAY_REFRACT		   ||
		     s->type == miPHOTON_TRANSMIT_SPECULAR ||
		     s->type == miPHOTON_TRANSMIT_GLOSSY   ||
		     s->type == miPHOTON_TRANSMIT_DIFFUSE) &&
		     s->parent && s->parent->shader == state->shader)
		{
			n++;
			if (!s_in)
				s_in = s->parent;
		}
	if (n & 1)					/* odd: exiting */
		state->refraction_volume = s_in ? s_in->volume : state->camera->volume;
	else						/* even: entering */
		if (!state->refraction_volume)
			state->refraction_volume = state->volume;
}


extern "C"
{
DLLEXPORT miBoolean p_CSG_init(miState *state, p_CSG_t *paras, miBoolean *inst_req)
{
	if (!paras)
	{
		*inst_req = miTRUE;
		return(miTRUE);
	}	
	if(state->options->shadow && state->options->shadow != 's')
		mi_warning("p_CSG shadow shader will produce correct shadows only with \"segments\" shadow mode!");
	return(miTRUE);
}

DLLEXPORT int p_CSG_version(void) {return(2);}

DLLEXPORT miBoolean p_CSG(miColor *result, miState *state, p_CSG_t *param)
{
	if(state->type == miRAY_DISPLACE)
		return miFALSE;
	if(state->type == miRAY_SHADOW)  	//shadow shader call
	{
		if(state->options->shadow == 's')	//segments shadow
		{
			result->r = result->g = result->b = 1;
			assert(state->instance);
			miInstance *inst = (miInstance *)mi_db_access(state->instance);
			const char *obj_name = inst->parent ? mi_api_tag_lookup(inst->parent) : 0;		//abject name
	//		const char *obj_name = inst->item ? mi_api_tag_lookup(inst->item) : 0;			//shape name
			mi_db_unpin(state->instance);

			miTag tag = 0;
			if(*mi_eval_boolean(&param->swap_AB))
				tag = *mi_eval_tag(&param->object_B);
			else
				tag = *mi_eval_tag(&param->object_A);
			if (!tag)
				return miFALSE;
			char *A = (char*)mi_db_access(tag);
			mi_db_unpin(tag);

			tag = 0;
			if(*mi_eval_boolean(&param->swap_AB))
				tag = *mi_eval_tag(&param->object_A);
			else
				tag = *mi_eval_tag(&param->object_B);
			if (!tag)
				return miFALSE;
			char *B = (char*)mi_db_access(tag);
			mi_db_unpin(tag);

			int mod = *mi_eval_integer(&param->mode);
			
			
			// first shadow call, we set "first_shd"
			int *first_shd;
			int cur_first_shd;
			int size_first_shd = 0;
			first_shd = (int *)mi_shaderstate_get(state, "first_shd", &size_first_shd);
			if (first_shd != NULL)
				cur_first_shd = 0;
			else //initialise it.
				cur_first_shd = 1;
		
			mi_shaderstate_set(state, "first_shd", &cur_first_shd, sizeof(cur_first_shd), miSS_LIFETIME_EYERAY);


			inside * rayStats;
			inside csg_mat;
			int size_inside = 0;
			rayStats = (inside *)mi_shaderstate_get(state, "csg_mat", &size_inside);
			if (rayStats != NULL)
				csg_mat = *rayStats;
			else //initialise it.
			{
				csg_mat.inside_a = miFALSE;
				csg_mat.inside_b = miFALSE;
				csg_mat.p_inside_a = miFALSE;		//previous object
				csg_mat.p_inside_b = miFALSE;		//previous object 

				//this need for making standart situation, if shader was called not from p_CSG
				if(cur_first_shd > 0)
				{
					if(strcmp(obj_name, A) == 0)
					{
						if(!mi_vector_dot(&state->dir, &state->parent->normal_geom))
							csg_mat.inside_a = miTRUE;
					}
					if(strcmp(obj_name, B) == 0)
					{
						if(!mi_vector_dot(&state->dir, &state->parent->normal_geom))
							csg_mat.inside_b = miTRUE;
					}
				}
			}
//			mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);

			inside_shd * ray_shdStats;
			inside_shd csg_shad;
			int size_shd_inside = 0;
			ray_shdStats = (inside_shd *)mi_shaderstate_get(state, "csg_shad", &size_shd_inside);
			if (ray_shdStats != NULL)
			{
				csg_shad = *ray_shdStats;		
				//copy previous value
				csg_shad.p_inside_a = csg_shad.inside_a;		
				csg_shad.p_inside_b = csg_shad.inside_b;
			}
			else //initialise it.
			{
				csg_shad.inside_a = miFALSE;
				csg_shad.inside_b = miFALSE;
				csg_shad.p_inside_a = miFALSE;
				csg_shad.p_inside_b = miFALSE;
			}
//			mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
			

			switch(mod)
			{
				case 0:				//union
					result->r = result->g = result->b = 0;
					return(miTRUE);
					break;
				case 1:				//difference
					if(cur_first_shd > 0)	// first shadow call, we read "csg_mat" (that come from material shader)and set "csg_shad" for read and write for follow shadow calls 
					{
						if(mi_vector_dot(&state->dir, &state->parent->normal_geom) && (csg_mat.inside_a == miTRUE || csg_mat.inside_b == miTRUE))	//we get incorrect "csg_mat" (opposite direction), just restore previous value
						{
							csg_mat.inside_a = csg_mat.p_inside_a;
							csg_mat.inside_b = csg_mat.p_inside_b;
						}

						if((csg_mat.inside_a == miTRUE) && (csg_mat.inside_b == miTRUE))		//insiade A & B
						{
							if(strcmp(obj_name, A) == 0)
							{
								if(state->inv_normal == miTRUE)
									csg_shad.inside_a = miFALSE;							
								else		//possible double A
									csg_shad.inside_a = miTRUE;
								
								mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
								result->r = result->g = result->b = 0;
							}
							else			//mean object B
							{
								if(strcmp(obj_name, B) == 0)
								{
									if(state->inv_normal == miTRUE)
										csg_shad.inside_b = miFALSE;
									else
										csg_shad.inside_b = miTRUE;
									mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
									mi_continue_shadow_seg(result, state);
								}
								else			//error, not defined object name
								{
									result->r = result->g = result->b = 0;
									return(miTRUE);
								}
							}
						}
						else			//inside A or B or None
						{
							if(csg_mat.inside_a == miTRUE)				//inside A
							{
								if(strcmp(obj_name, A) == 0)
								{
									if(state->inv_normal == miTRUE)
										csg_shad.inside_a = miFALSE;
									else
										csg_shad.inside_a = miTRUE;
									mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
									mi_continue_shadow_seg(result, state);
								}
								else		//mean B
								{
									if(state->inv_normal == miTRUE)
										csg_shad.inside_b = miFALSE;
									else
										csg_shad.inside_b = miTRUE;
									mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
									mi_continue_shadow_seg(result, state);
								}
							}
							else
							{
								if(csg_mat.inside_b == miTRUE)				//inside B
								{
									if(strcmp(obj_name, B) == 0)
									{
										if(state->inv_normal == miTRUE)
											csg_shad.inside_b = miFALSE;
										else
											csg_shad.inside_b = miTRUE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										result->r = result->g = result->b = 0;
									}
									else		//mean A
									{
										if(state->inv_normal == miTRUE)
											csg_shad.inside_a = miFALSE;
										else
											csg_shad.inside_a = miTRUE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										result->r = result->g = result->b = 0;
									}
								}
								else								//inside None
								{
									if(strcmp(obj_name, A) == 0)
									{									
										if(state->inv_normal == miFALSE)
											csg_shad.inside_a = miTRUE;
										else
											csg_shad.inside_a = miFALSE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										mi_continue_shadow_seg(result, state);
									}
									else		//mean B or error
									{
										if(state->inv_normal == miFALSE)
											csg_shad.inside_b = miTRUE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										result->r = result->g = result->b = 0;
									}
								}
							}
						}
					}
					else //it secondary shadow call, we use "csg_shad"
					{
						if((csg_shad.inside_a == miTRUE) && (csg_shad.inside_b == miTRUE))		//insiade A & B
						{
							if(strcmp(obj_name, A) == 0)
							{
								if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
									csg_shad.inside_a = miFALSE;							
								else		//possible double A
									csg_shad.inside_a = miTRUE;
								
								mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
								result->r = result->g = result->b = 0;
							}
							else			//mean object B
							{
								if(strcmp(obj_name, B) == 0)
								{	
									if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
										csg_shad.inside_b = miFALSE;
									else
										csg_shad.inside_b = miTRUE;

										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
									mi_continue_shadow_seg(result, state);
								}
								else			//error, not defined object name
								{
									result->r = result->g = result->b = 0;
									return(miTRUE);
								}
							}
						}
						else			//inside A or B or None
						{
							if(csg_shad.inside_a == miTRUE)				//inside A
							{
								if(strcmp(obj_name, A) == 0)
								{
									if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
										csg_shad.inside_a = miFALSE;
									else
										csg_shad.inside_a = miTRUE;
									mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
									mi_continue_shadow_seg(result, state);
								}
								else		//mean B
								{
									if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
										csg_shad.inside_b = miFALSE;
									else
										csg_shad.inside_b = miTRUE;
									mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
									mi_continue_shadow_seg(result, state);
								}
							}
							else
							{
								if(csg_shad.inside_b == miTRUE)				//inside B
								{
									if(strcmp(obj_name, B) == 0)
									{
										if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
											csg_shad.inside_b = miFALSE;
										else
											csg_shad.inside_b = miTRUE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										result->r = result->g = result->b = 0;
									}
									else		//mean A
									{
										if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
											csg_shad.inside_a = miFALSE;
										else
											csg_shad.inside_a = miTRUE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										result->r = result->g = result->b = 0;
									}
								}
								else								//inside None
								{
									if(strcmp(obj_name, A) == 0)
									{
										if(!mi_vector_dot(&state->dir, &state->parent->normal_geom))
											csg_shad.inside_a = miTRUE;
										else
											csg_shad.inside_a = miFALSE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										mi_continue_shadow_seg(result, state);
									}
									else		//mean B
									{
										if(!mi_vector_dot(&state->dir, &state->parent->normal_geom))
											csg_shad.inside_b = miTRUE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										result->r = result->g = result->b = 0;
									}
								}
							}
						}
					}
					break;
				////////////////////////////////////////////////////////////
				case 2:				//intersection
					if(cur_first_shd > 0)	// first shadow call, we read "csg_mat" (that come from material shader)and set "csg_shad" for read and write for follow shadow calls 
					{
						if(mi_vector_dot(&state->dir, &state->parent->normal_geom) && (csg_mat.inside_a == miTRUE || csg_mat.inside_b == miTRUE))	//we get incorrect "csg_mat" (opposite direction), just restore previous value
						{
							csg_mat.inside_a = csg_mat.p_inside_a;
							csg_mat.inside_b = csg_mat.p_inside_b;
						}

						if((csg_mat.inside_a == miTRUE) && (csg_mat.inside_b == miTRUE))		//insiade A & B
						{
							if(strcmp(obj_name, A) == 0)
							{
								if(state->inv_normal == miTRUE)
									csg_shad.inside_a = miFALSE;							
								else		//possible double A
									csg_shad.inside_a = miTRUE;
								
								mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
								result->r = result->g = result->b = 0;
							}
							else			//mean object B
							{
								if(strcmp(obj_name, B) == 0)
								{
									if(state->inv_normal == miTRUE)
										csg_shad.inside_b = miFALSE;
									else
										csg_shad.inside_b = miTRUE;
									mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
									result->r = result->g = result->b = 0;
								}
								else			//error, not defined object name
								{
									result->r = result->g = result->b = 0;
									return(miTRUE);
								}
							}
						}
						else			//inside A or B or None
						{
							if(csg_mat.inside_a == miTRUE)				//inside A
							{
								if(strcmp(obj_name, A) == 0)
								{
									if(state->inv_normal == miTRUE)
										csg_shad.inside_a = miFALSE;
									else
										csg_shad.inside_a = miTRUE;
									mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
									mi_continue_shadow_seg(result, state);
								}
								else		//mean B
								{
									if(state->inv_normal == miTRUE)
										csg_shad.inside_b = miFALSE;
									else
										csg_shad.inside_b = miTRUE;
									mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
									result->r = result->g = result->b = 0;
								}
							}
							else
							{
								if(csg_mat.inside_b == miTRUE)				//inside B
								{
									if(strcmp(obj_name, B) == 0)
									{
										if(state->inv_normal == miTRUE)
											csg_shad.inside_b = miFALSE;
										else
											csg_shad.inside_b = miTRUE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										mi_continue_shadow_seg(result, state);
									}
									else		//mean A
									{
										if(state->inv_normal == miTRUE)
											csg_shad.inside_a = miFALSE;
										else
											csg_shad.inside_a = miTRUE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										result->r = result->g = result->b = 0;
									}
								}
								else								//inside None
								{
									if(strcmp(obj_name, A) == 0)
									{									
										if(state->inv_normal == miFALSE)
											csg_shad.inside_a = miTRUE;
										else
											csg_shad.inside_a = miFALSE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										mi_continue_shadow_seg(result, state);
									}
									else		//mean B or error
									{
										if(state->inv_normal == miFALSE)
											csg_shad.inside_b = miTRUE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										mi_continue_shadow_seg(result, state);
									}
								}
							}
						}
					}
					else //it secondary shadow call, we use "csg_shad"
					{
						if((csg_shad.inside_a == miTRUE) && (csg_shad.inside_b == miTRUE))		//insiade A & B
						{
							if(strcmp(obj_name, A) == 0)
							{
								if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
									csg_shad.inside_a = miFALSE;							
								else		//possible double A
									csg_shad.inside_a = miTRUE;
								
								mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
								result->r = result->g = result->b = 0;
							}
							else			//mean object B
							{
								if(strcmp(obj_name, B) == 0)
								{	
									if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
										csg_shad.inside_b = miFALSE;
									else
										csg_shad.inside_b = miTRUE;

									mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
									result->r = result->g = result->b = 0;
								}
								else			//error, not defined object name
								{
									result->r = result->g = result->b = 0;
									return(miTRUE);
								}
							}
						}
						else			//inside A or B or None
						{
							if(csg_shad.inside_a == miTRUE)				//inside A
							{
								if(strcmp(obj_name, A) == 0)
								{
									if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
										csg_shad.inside_a = miFALSE;
									else
										csg_shad.inside_a = miTRUE;
									mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
									mi_continue_shadow_seg(result, state);
								}
								else		//mean B
								{
									if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
										csg_shad.inside_b = miFALSE;
									else
										csg_shad.inside_b = miTRUE;
									mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
									result->r = result->g = result->b = 0;
								}
							}
							else
							{
								if(csg_shad.inside_b == miTRUE)				//inside B
								{
									if(strcmp(obj_name, B) == 0)
									{
										if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
											csg_shad.inside_b = miFALSE;
										else
											csg_shad.inside_b = miTRUE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										mi_continue_shadow_seg(result, state);
									}
									else		//mean A
									{
										if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
											csg_shad.inside_a = miFALSE;
										else
											csg_shad.inside_a = miTRUE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										result->r = result->g = result->b = 0;
									}
								}
								else								//inside None
								{
									if(strcmp(obj_name, A) == 0)
									{
										if(!mi_vector_dot(&state->dir, &state->parent->normal_geom))
											csg_shad.inside_a = miTRUE;
										else
											csg_shad.inside_a = miFALSE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										mi_continue_shadow_seg(result, state);
									}
									else		//mean B
									{
										if(!mi_vector_dot(&state->dir, &state->parent->normal_geom))
											csg_shad.inside_b = miTRUE;
										mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
										mi_continue_shadow_seg(result, state);
									}
								}
							}
						}
					}
					break;
				default:
					result->r = result->g = result->b = 0;
					break;
			}
			if(cur_first_shd == 1)		//cleanup
			{
				mi_shaderstate_set(state, "first_shd", NULL, sizeof(NULL), miSS_LIFETIME_EYERAY);
				mi_shaderstate_set(state, "csg_shad", NULL, sizeof(NULL), miSS_LIFETIME_EYERAY);
			}
		}	
		else       //else segments shadow
		{
			result->r = result->g = result->b = result->a = 0;
			return(miTRUE);
		}
	}
	else		//else shadow
	{
		assert(state->instance);
		miInstance *inst = (miInstance *)mi_db_access(state->instance);
		const char *obj_name = inst->parent ? mi_api_tag_lookup(inst->parent) : 0;
//		const char *obj_name = inst->item ? mi_api_tag_lookup(inst->item) : 0;
		mi_db_unpin(state->instance);

		miTag tag = 0;
		if(*mi_eval_boolean(&param->swap_AB))
			tag = *mi_eval_tag(&param->object_B);
		else
			tag = *mi_eval_tag(&param->object_A);
		if (!tag)
			return miFALSE;
		char *A = (char*)mi_db_access(tag);
		mi_db_unpin(tag);

		tag = 0;
		if(*mi_eval_boolean(&param->swap_AB))
			tag = *mi_eval_tag(&param->object_A);
		else
			tag = *mi_eval_tag(&param->object_B);
		if (!tag)
			return miFALSE;
		char *B = (char*)mi_db_access(tag);
		mi_db_unpin(tag);

		int mod = *mi_eval_integer(&param->mode);

		inside * rayStats;
		inside csg_mat;
		int size_inside = 0;

		rayStats = (inside *)mi_shaderstate_get(state, "csg_mat", &size_inside);
		if (rayStats != NULL)
		{
			csg_mat = *rayStats;
			//copy previous value
			csg_mat.p_inside_a = csg_mat.inside_a;		
			csg_mat.p_inside_b = csg_mat.inside_b;
		}
		else //initialise it.
		{
			csg_mat.inside_a = miFALSE;
			csg_mat.inside_b = miFALSE;
			csg_mat.p_inside_a = miFALSE;
			csg_mat.p_inside_b = miFALSE;
		}

		if(*mi_eval_boolean(&param->check_camera) && state->type == miRAY_EYE)
		{
			miVector inv_dir = state->dir; 
			mi_vector_neg(&inv_dir);
			if(mi_trace_probe(state, &inv_dir, &state->point))
			{
				if(state->child->instance)
				{
					miInstance *inv_inst = (miInstance *)mi_db_access(state->child->instance);
					const char *inv_obj_name = inv_inst->parent ? mi_api_tag_lookup(inv_inst->parent) : 0;
			//		const char *inv_obj_name = inv_inst->item ? mi_api_tag_lookup(inst->item) : 0;
					mi_db_unpin(state->child->instance);
					
					if(strcmp(inv_obj_name, A) == 0)
					{
						if(state->child->inv_normal)
							csg_mat.inside_a = miTRUE;
					}
					else if (strcmp(inv_obj_name, B) == 0)
					{
						if(state->child->inv_normal)
							csg_mat.inside_b = miTRUE;
					}
				}
			}
		}
	

		switch(mod)
		{
			case 0:				//union
				if((csg_mat.inside_a == miTRUE) && (csg_mat.inside_b == miTRUE))		//insiade A & B
				{
					if(strcmp(obj_name, A) == 0)
					{
						if(state->inv_normal == miTRUE)
							csg_mat.inside_a = miFALSE;
						else
							csg_mat.inside_a = miTRUE;
					}
					else
					{
						if(strcmp(obj_name, B) == 0)
						{
							if(state->inv_normal == miTRUE)
								csg_mat.inside_b = miFALSE;
							else
								csg_mat.inside_b = miTRUE;
						}
						else			//error
						{
							result->r = 1;
							result->g = result->b = 0;
							return(miTRUE);
						}
					}
					mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
					mi_trace_continue(result, state);
				}
				else
				{
					if(csg_mat.inside_a == miTRUE)				//inside A
					{
						if(strcmp(obj_name, A) == 0)
						{
							if(state->inv_normal == miTRUE)
								csg_mat.inside_a = miFALSE;
							else
								csg_mat.inside_a = miTRUE;
							mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
							*result = *mi_eval_color(&param->color_A);
						}
						else		//mean B
						{
							if(state->inv_normal == miTRUE)
								csg_mat.inside_b = miFALSE;
							else
								csg_mat.inside_b = miTRUE;
							mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
							mi_trace_continue(result, state);
						}
					}
					else
					{
						if(csg_mat.inside_b == miTRUE)				//inside B
						{
							if(strcmp(obj_name, B) == 0)
							{
								if(state->inv_normal == miTRUE)
									csg_mat.inside_b = miFALSE;
								else
									csg_mat.inside_b = miTRUE;
								mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
								*result = *mi_eval_color(&param->color_B);
							}
							else		//mean A
							{
								if(state->inv_normal == miTRUE)
									csg_mat.inside_a = miFALSE;
								else
									csg_mat.inside_a = miTRUE;
								mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
								mi_trace_continue(result, state);
							}
						}
						else
						{
							if(strcmp(obj_name, A) == 0)
							{
								if(state->inv_normal == miFALSE)
									csg_mat.inside_a = miTRUE;
								mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
								*result = *mi_eval_color(&param->color_A);
							}
							else
							{
								if(state->inv_normal == miFALSE)
									csg_mat.inside_b = miTRUE;
								mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
								*result = *mi_eval_color(&param->color_B);
							}
						}
					}
				}
				break;
			case 1:				//difference
				if((csg_mat.inside_a == miTRUE) && (csg_mat.inside_b == miTRUE))		//insiade A & B
				{
					if(strcmp(obj_name, A) == 0)
					{
						if(state->inv_normal == miTRUE)
							csg_mat.inside_a = miFALSE;
						else
							csg_mat.inside_a = miTRUE;
						mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
						if(*mi_eval_boolean(&param->swap_AB))
							*result = *mi_eval_color(&param->color_B);
						else
							*result = *mi_eval_color(&param->color_A);
					}
					else
					{
						if(strcmp(obj_name, B) == 0)
						{
							if(state->inv_normal == miTRUE)
								csg_mat.inside_b = miFALSE;
							else
								csg_mat.inside_b = miTRUE;
							mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
							mi_trace_continue(result, state);
						}
						else			//error
						{
							result->r = 1;
							result->g = result->b = 0;
							return(miTRUE);
						}
					}
				}
				else			//inside A or B or None
				{
					if(csg_mat.inside_a == miTRUE)				//inside A
					{
						if(strcmp(obj_name, A) == 0)
						{
							if(state->inv_normal == miTRUE)
								csg_mat.inside_a = miFALSE;
							else
								csg_mat.inside_a = miTRUE;
							mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
							mi_trace_continue(result, state);
						}
						else		//mean B
						{
							if(state->inv_normal == miTRUE)
								csg_mat.inside_b = miFALSE;
							else
								csg_mat.inside_b = miTRUE;
							mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
							mi_trace_continue(result, state);
						}
					}
					else
					{
						if(csg_mat.inside_b == miTRUE)				//inside B
						{
							if(strcmp(obj_name, B) == 0)
							{
								if(state->inv_normal == miTRUE)
									csg_mat.inside_b = miFALSE;
								else
									csg_mat.inside_b = miTRUE;
								mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
								if(*mi_eval_boolean(&param->swap_AB))
									*result = *mi_eval_color(&param->color_A);
								else
									*result = *mi_eval_color(&param->color_B);
							}
							else		//mean A
							{
								if(state->inv_normal == miTRUE)
									csg_mat.inside_a = miFALSE;
								else
									csg_mat.inside_a = miTRUE;
								mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
								if(*mi_eval_boolean(&param->swap_AB))
									*result = *mi_eval_color(&param->color_B);
								else
									*result = *mi_eval_color(&param->color_A);
							}
						}
						else					//inside None
						{
							if(strcmp(obj_name, A) == 0)
							{
								if(state->inv_normal == miFALSE)
									csg_mat.inside_a = miTRUE;
								mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
								mi_trace_continue(result, state);
							}
							else
							{
								if(state->inv_normal == miFALSE)
									csg_mat.inside_b = miTRUE;
								mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
								if(*mi_eval_boolean(&param->swap_AB))
									*result = *mi_eval_color(&param->color_A);
								else
									*result = *mi_eval_color(&param->color_B);
							}
						}
					}
				}
				break;
			case 2:				//intersection
				if((csg_mat.inside_a == miTRUE) && (csg_mat.inside_b == miTRUE))		//insiade A & B
				{
					if(strcmp(obj_name, A) == 0)
					{
						if(state->inv_normal == miTRUE)
							csg_mat.inside_a = miFALSE;
						else
							csg_mat.inside_a = miTRUE;
						mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
						*result = *mi_eval_color(&param->color_A);
					}
					else
					{
						if(strcmp(obj_name, B) == 0)
						{
							if(state->inv_normal == miTRUE)
								csg_mat.inside_b = miFALSE;
							else
								csg_mat.inside_b = miTRUE;
							mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
							*result = *mi_eval_color(&param->color_B);
						}
						else			//error
						{
							result->r = 1;
							result->g = result->b = 0;
							return(miTRUE);
						}
					}
				}
				else
				{
					if(csg_mat.inside_a == miTRUE)				//inside A
					{
						if(strcmp(obj_name, A) == 0)
						{
							if(state->inv_normal == miTRUE)
								csg_mat.inside_a = miFALSE;
							else
								csg_mat.inside_a = miTRUE;
							mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
							mi_trace_continue(result, state);
						}
						else		//mean B
						{
							if(state->inv_normal == miTRUE)
								csg_mat.inside_b = miFALSE;
							else
								csg_mat.inside_b = miTRUE;
							mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
							*result = *mi_eval_color(&param->color_B);
						}
					}
					else
					{
						if(csg_mat.inside_b == miTRUE)				//inside B
						{
							if(strcmp(obj_name, B) == 0)
							{
								if(state->inv_normal == miTRUE)
									csg_mat.inside_b = miFALSE;
								else
									csg_mat.inside_b = miTRUE;
								mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
								mi_trace_continue(result, state);
							}
							else		//mean A
							{
								if(state->inv_normal == miTRUE)
									csg_mat.inside_a = miFALSE;
								else
									csg_mat.inside_a = miTRUE;
								mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
								*result = *mi_eval_color(&param->color_A);
							}
						}
						else
						{
							if(strcmp(obj_name, A) == 0)
							{
								if(state->inv_normal == miFALSE)
									csg_mat.inside_a = miTRUE;
								else
									csg_mat.inside_a = miFALSE;
								mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
								mi_trace_continue(result, state);
							}
							else
							{
								if(state->inv_normal == miFALSE)
									csg_mat.inside_b = miTRUE;
								else
									csg_mat.inside_b = miFALSE;
								mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
								mi_trace_continue(result, state);
							}
						}
					}
				}
				break;
			default:
				result->r = 1;
				result->g = result->b = 0;
				break;
		}
		mi_shaderstate_set(state, "csg_mat", NULL, sizeof(NULL), miSS_LIFETIME_EYERAY);
	}
	return(miTRUE);
}
}
