//  Copyright (c) 2006, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"
#include "geoshader.h"
#include "p_utils.h"

#define EPS	 1e-4
#define BLACK(C) ((C).r==0 && (C).g==0 && (C).b==0)


typedef struct
{
	miColor         color;				// color of light source
	miScalar		intensity;
	miBoolean		emit_diffuse;
	miBoolean		emit_specular;
	int			    atten;				// distance attenuation
	miScalar        start, stop;		// if atten, distance range	
	miScalar        cone;				// inner solid cone
	miScalar		cone_falloff;		// cone gamma
	miBoolean       shadow;				// light casts shadows
	miScalar        transparency;       // makes opaque objects transparent
	miScalar		light_radius;		// shadow rays spread
	int				shadow_rays;		// shadow samples
	miScalar		threshold;			// shadow threshold
	miTag			light_category;
}p_SpotTK_t;


typedef struct		//struct that p_LightTK return
{
	miColor l_color_shd;		//light color with shadow
	miColor	l_color;			//light color
	miBoolean emit_diffuse;
	miBoolean emit_specular;
	char category[32];
}light_color;

extern "C"
{
DLLEXPORT int p_SpotTK_version(void) {return(2);}

DLLEXPORT miBoolean p_SpotTK(register miColor *result, register miState *state, register p_SpotTK_t *paras)
{
	register miScalar       d, transp, t, start, stop, cone;
	miScalar                spread;
	miScalar				intens;
	miVector                ldir, dir;
	miTag                   ltag;
	miScalar		cone_falloff;
	miScalar		light_radius;
	miColor			filter;
	miColor			sum_result;
	miColor			cur_result;
	miVector		shiftVL;
	double			sample[2];
	int				sample_count = 0;
	miUint			shadow_rays;
	int TK = 1;
	light_color lightTK;
	int passive = 0;			//if it set to 1, shader compute without shadow, only for light pass and return miTRUE
//	else
//		mi_warning("%s", "Maya Shaders...");
	intens = *mi_eval_scalar(&paras->intensity);
	*result = *mi_eval_color(&paras->color);
	result->r *= intens;
	result->g *= intens;
	result->b *= intens;
	if (state->type != miRAY_LIGHT)                 // visible area light
		return(miTRUE);
                                                               
	ltag = ((miInstance *)mi_db_access(state->light_instance))->item;
	mi_db_unpin(state->light_instance);
	mi_query(miQ_LIGHT_DIRECTION, state, ltag, &ldir);
	mi_vector_to_light(state, &dir, &state->dir);
	mi_vector_normalize(&dir);
	d = mi_vector_dot(&dir, &ldir);

	miColor temp_light_color;
	temp_light_color = *result;
	if (d <= 0)
	{
		if(TK)
		{
			result->r = result->g = result->b = 0;
			passive = 1;		//ignore shadow for followed computing, but we need light color
		}
		else
			return(miFALSE);
	}
	mi_query(miQ_LIGHT_SPREAD, state, ltag, &spread);
	
	if(d <= spread)
	{
		if(TK)
		{
			result->r = result->g = result->b = 0;
			passive = 1;		//ignore shadow for followed computing, but we need light color
		}
		else
			return(miFALSE);
	}
	cone = *mi_eval_scalar(&paras->cone);
	cone = (miScalar)(cone * ATR)/2;
	cone = cos(cone);
	if (d < cone && !passive)
	{
		t = 1 - (d - cone) / (spread - cone);
		if(t > 0 && t < 1)
		{
			cone_falloff = *mi_eval_scalar(&paras->cone_falloff);
			if(cone_falloff != 1)
				t = pow(t, cone_falloff);
		}
		//clamp value
		if(t < 0)		
			t = 0;
		else if (t > 1)
			t = 1;
		result->r *= t;
		result->g *= t;
		result->b *= t;
	}

	int atten = *mi_eval_integer(&paras->atten);
	if(atten > 0)	//atten
	{
		double distance;
		start = *mi_eval_scalar(&paras->start);
		stop = *mi_eval_scalar(&paras->stop);
		distance = state->dist;
		switch (atten)
		{
			case 1:			//distance linear
				if (distance >= stop)
				{
					if(TK)
					{
						result->r = result->g = result->b = 0;
						temp_light_color.r = temp_light_color.g = temp_light_color.b = 0;
						break;
					}
					else
						return(miFALSE);
				}

				if (distance > start && fabs(stop - start) > EPS)
				{
					t = (miScalar)(1 - (distance - start) / (stop - start));
					result->r *= t;
					result->g *= t;
					result->b *= t;
					if(TK)
					{
						temp_light_color.r *= t;
						temp_light_color.g *= t;
						temp_light_color.b *= t;
					}
				}
				break;
			case 2:				//linear
				if(distance > start)
				{
					if(start < 1)
					{
						start++;
						distance++;
					}
					miScalar inv_dist = (miScalar)(1.f/(distance -start +1));
					result->r *= inv_dist;
					result->g *= inv_dist;
					result->b *= inv_dist;
					if(TK)
					{
						temp_light_color.r *= inv_dist;
						temp_light_color.g *= inv_dist;
						temp_light_color.b *= inv_dist;
					}
				}
				break;
			case 3:				//quadratic distance
				if(distance > start)
				{
					if(start < 1)
					{
						start++;
						stop++;
						distance++;
					}
					distance = distance -start +1;
					distance *= distance;
					distance--;
					distance /= stop - start;	//map to distance
					distance++;
					miScalar dist_qudr = 1.f/(miScalar)distance;
					result->r *= dist_qudr;
					result->g *= dist_qudr;
					result->b *= dist_qudr;
					if(TK)
					{
						temp_light_color.r *= dist_qudr;
						temp_light_color.g *= dist_qudr;
						temp_light_color.b *= dist_qudr;
					}
				}
				break;
			case 4:				//quadratic
				if(distance > start)
				{
					if(start < 1)
					{
						start++;
						distance++;
					}
					distance = distance -start +1;
					miScalar dist_qudr = 1.f/(miScalar)(distance * distance);
					result->r *= dist_qudr;
					result->g *= dist_qudr;
					result->b *= dist_qudr;
					if(TK)
					{
						temp_light_color.r *= dist_qudr;
						temp_light_color.g *= dist_qudr;
						temp_light_color.b *= dist_qudr;
					}
				}
				break;
			case 5:				//cubic distance
				if(distance > start)
				{
					if(start < 1)
					{
						start++;
						stop++;
						distance++;
					}
					distance = distance -start +1;
					distance = pow(distance, 3);
					distance--;
					distance /= stop - start;	//map to distance
					distance++;
					miScalar dist_qudr = 1.f/(miScalar)distance;
					result->r *= dist_qudr;
					result->g *= dist_qudr;
					result->b *= dist_qudr;
					if(TK)
					{
						temp_light_color.r *= dist_qudr;
						temp_light_color.g *= dist_qudr;
						temp_light_color.b *= dist_qudr;
					}
				}
				break;
			case 6:				//cubic
				if(distance > start)
				{
					if(start < 1)
					{
						start++;
						distance++;
					}
					distance = distance -start +1;
					distance = pow(distance, 3);
					miScalar dist_qudr = 1.f/(miScalar)distance;
					result->r *= dist_qudr;
					result->g *= dist_qudr;
					result->b *= dist_qudr;
					if(TK)
					{
						temp_light_color.r *= dist_qudr;
						temp_light_color.g *= dist_qudr;
						temp_light_color.b *= dist_qudr;
					}
				}
				break;
			default:
				break;
		}
	}

	miScalar threshold = *mi_eval_scalar(&paras->threshold);
	if(*mi_eval_boolean(&paras->shadow) && !passive && (result->r > threshold || result->g > threshold || result->b > threshold))  // shadows
	{
		transp = *mi_eval_scalar(&paras->transparency);
		if (transp < 1)
		{			
			filter.r = filter.g = filter.b = filter.a = 1;			
			light_radius = *mi_eval_scalar(&paras->light_radius);
            			
			cur_result = *result;
			sum_result.r = sum_result.g = sum_result.b = sum_result.a = 0;

			miVector light_dir;
			miVector light_org = state->org;
			miVector orig_light_org = state->org;		//save original light position
			miVector original_ldir = state->dir;		//save original direction to light
			             
			if(light_radius > 0)
			{
				//decrease shadow rays by reflect/refract depth
				shadow_rays = *mi_eval_integer(&paras->shadow_rays);
				if(shadow_rays > 1)
				{
					if(final_gather_mode(state))
						shadow_rays = 1;
					else
					{
						int level = state->refraction_level + state->reflection_level;
						shadow_rays >>= level;
						if(shadow_rays < 1)
							shadow_rays = 1;
					}
				}

				double inv_transp = 1 - transp;
				while(mi_sample(sample, &sample_count, state, 2, &shadow_rays)) 
				{
					shiftVL = p_sphere_sample(sample);
					mi_vector_mul(&shiftVL, light_radius);

					mi_vector_add(&light_org, &light_org, &shiftVL);		//mix random vector and light direction
					mi_vector_sub(&light_dir, &light_org, &state->point);
					mi_vector_neg(&light_dir);
					mi_vector_normalize(&light_dir);

					filter.r = filter.g = filter.b = filter.a = 1;
					state->dir = light_dir;
					state->org = light_org;

					if(!mi_trace_shadow(&filter, state) || BLACK(filter))	//black shadow
					{
						result->r *= transp;
						result->g *= transp;
						result->b *= transp;
					}
					else // transparent
					{
						result->r *= (float)(transp + inv_transp * filter.r);
						result->g *= (float)(transp + inv_transp * filter.g);
						result->b *= (float)(transp + inv_transp * filter.b);
					}
					sum_result.r += result->r / shadow_rays;
					sum_result.g += result->g / shadow_rays;
					sum_result.b += result->b / shadow_rays;
					*result = cur_result;
					light_org = orig_light_org;
					state->dir = original_ldir;
					state->org = orig_light_org;
				}
				*result = sum_result;
			}
			else
			{
				if(!mi_trace_shadow(&filter, state) || BLACK(filter))
				{
					result->r *= transp; 
					result->g *= transp;
					result->b *= transp;
					if (transp == 0 && !TK)
							return(miFALSE);
				}
				else // transparent
				{
					double inv_transp = 1 - transp;
					result->r *= (float)(transp + inv_transp * filter.r);
					result->g *= (float)(transp + inv_transp * filter.g);
					result->b *= (float)(transp + inv_transp * filter.b);
				}
			}
			state->dir = original_ldir;		//restore original light direction
			state->org = orig_light_org;	//restore original light position
		}
	}
	if(TK)
	{
		lightTK.emit_diffuse = *mi_eval_boolean(&paras->emit_diffuse);
		lightTK.emit_specular = *mi_eval_boolean(&paras->emit_specular);
		lightTK.l_color = temp_light_color;
		lightTK.l_color_shd = *result;

		char *category = lightTK.category;
		char *eval_categ;
		miTag tag;
		if(paras->light_category)
		{
			tag = *mi_eval_tag(&paras->light_category);
			eval_categ = (char *)mi_db_access(tag);
			mi_db_unpin(tag);
			if(!eval_categ || !eval_categ[0])
			{
//				mi_warning("p_SpotTK: empty light category specified.");
				for(int i = 0; i < 32; i++)
					category[i] = 0;
			}
			else
			{
				int counter = 0;
				while(*eval_categ || (counter < 31))
				{
					*category = *eval_categ;
					eval_categ++;
					category++;
					counter++;
				}
				category = NULL;
			}
		}
		else
			mi_warning("p_SpotTK: no light category specified.");

		mi_shaderstate_set(state, lightsTK_key, &lightTK, sizeof(lightTK), miSS_LIFETIME_EYERAY);
	}
	return(miTRUE);
}
}
