//  Copyright (c) 2005, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "shader.h"
#include "geoshader.h"

typedef struct
{
	miColor	glass_mat;
	miColor	liquid_mat;
	miColor	interface_mat;

	int object_mode;

	miTag glass_name;
	miTag liquid_name;

	miTag glass_object;
	miTag liquid_object;

	int glass_label;
	int liquid_label;

	miBoolean check_camera;
}p_dielectric_interface_t;

typedef struct
{
	miBoolean inside_glass;			//current glass value
	miBoolean inside_liquid;		//current liquid value
	miBoolean p_inside_glass;		//previous value of A, we can restore if need (mostly used for shadow shader)
	miBoolean p_inside_liquid;		//previous value of B, we can restore if need (mostly used for shadow shader)
	miBoolean p1_inside_glass;		//previous value of A, we can restore if need (mostly used for shadow shader)
	miBoolean p1_inside_liquid;		//previous value of B, we can restore if need (mostly used for shadow shader)
	miBoolean p2_inside_glass;		//previous value of A, we can restore if need (mostly used for shadow shader)
	miBoolean p2_inside_liquid;		//previous value of B, we can restore if need (mostly used for shadow shader)
	miBoolean p3_inside_glass;		//previous value of A, we can restore if need (mostly used for shadow shader)
	miBoolean p3_inside_liquid;		//previous value of B, we can restore if need (mostly used for shadow shader)
}inside;

typedef struct
{
	miBoolean inside_glass;
	miBoolean inside_liquid;
	miBoolean p_inside_glass;
	miBoolean p_inside_liquid;
}inside_shd;

static void choose_volume(miState *state)
{
	register int		n = 0;		/* same-material counter */
	register miState	*s, *s_in = 0;	/* for finding enclosing mtl */

	for (n=0, s=state; s; s=s->parent)		/* history? */
		if ((s->type == miRAY_TRANSPARENT	   ||
		     s->type == miRAY_REFRACT		   ||
		     s->type == miPHOTON_TRANSMIT_SPECULAR ||
		     s->type == miPHOTON_TRANSMIT_GLOSSY   ||
		     s->type == miPHOTON_TRANSMIT_DIFFUSE) &&
		     s->parent && s->parent->shader == state->shader)
		{
			n++;
			if (!s_in)
				s_in = s->parent;
		}
	if (n & 1)					/* odd: exiting */
		state->refraction_volume = s_in ? s_in->volume : state->camera->volume;
	else						/* even: entering */
		if (!state->refraction_volume)
			state->refraction_volume = state->volume;
}

static int p_get_label(miTag instance)
{
    int result       = 0;
    miInstance* inst = NULL;
    
    if (!instance) return 0;
    inst = (miInstance *)mi_db_access(instance);
    if (inst) result = inst->label;
    mi_db_unpin(instance);
    return result;
}

int current_object(miState *state, p_dielectric_interface_t *param)
{

	int object_mode = *mi_eval_integer(&param->object_mode);

	switch(object_mode)
	{ 
		case 0:				//by tag
			break;
		case 1:				//by object name
		{
			assert(state->instance);
			miInstance *inst = (miInstance *)mi_db_access(state->instance);
			const char *obj_name = inst->parent ? mi_api_tag_lookup(inst->parent) : 0;
	//		const char *obj_name = inst->item ? mi_api_tag_lookup(inst->item) : 0;
			mi_db_unpin(state->instance);

			miTag tag = miNULLTAG;
			tag = *mi_eval_tag(&param->glass_name);
			if (!tag)
				break;
			char *glass_name = (char*)mi_db_access(tag);
			mi_db_unpin(tag);

			tag = miNULLTAG;
			tag = *mi_eval_tag(&param->liquid_name);
			if (!tag)
				break;
			char *liquid_name = (char*)mi_db_access(tag);
			mi_db_unpin(tag);

			if(strcmp(obj_name, glass_name) == 0)
				return 1;
			else if (strcmp(obj_name, liquid_name) == 0)
				return 2;
			break;
		}
		case 2:				//by label
		{
			int curent_inst_label = p_get_label(state->instance);
			if(curent_inst_label)
			{
				int glass_label = *mi_eval_integer(&param->glass_label);
				int liquid_label = *mi_eval_integer(&param->liquid_label);
				if(curent_inst_label == glass_label)
					return 1;
				else if (curent_inst_label == liquid_label)
					return 2;
			}
			break;
		}
		default:
			break;
	}
	return 0;  //None
}


extern "C"
{
DLLEXPORT miBoolean p_dielectric_interface_init(miState *state, p_dielectric_interface_t *param, miBoolean *inst_req)
{
	if (!param)
	{
		*inst_req = miTRUE;
		return(miTRUE);
	}	
	if(state->options->shadow && state->options->shadow != 's')
		mi_warning("p_dielectric_interface shadow shader will produce correct shadows only with \"segments\" shadow mode!");
	return(miTRUE);
}

DLLEXPORT int p_dielectric_interface_version(void) {return(2);}

DLLEXPORT miBoolean p_dielectric_interface(miColor *result, miState *state, p_dielectric_interface_t *param)
{
	if(state->type == miRAY_DISPLACE)
		return miFALSE;

	int cur_object = current_object(state, param);		//check what object hitted

	if(state->type == miRAY_SHADOW)  	//shadow shader call
	{
		if(state->options->shadow == 's')	//segments shadow
		{
			result->r = result->g = result->b = 1;
		
			// first shadow call, we set "first_shd"
			int *first_shd;
			int cur_first_shd;
			first_shd = (int *)mi_shaderstate_get(state, "first_shd", NULL);
			if (first_shd != NULL)
				cur_first_shd = 0;
			else //initialise it.
				cur_first_shd = 1;
		
			mi_shaderstate_set(state, "first_shd", &cur_first_shd, sizeof(cur_first_shd), miSS_LIFETIME_EYERAY);


			inside * rayStats;
			inside csg_mat;
			rayStats = (inside *)mi_shaderstate_get(state, "csg_mat", NULL);
			if (rayStats != NULL)
				csg_mat = *rayStats;
			else //initialise it.
			{
				csg_mat.inside_glass = miFALSE;
				csg_mat.inside_liquid = miFALSE;
				csg_mat.p_inside_glass = miFALSE;		//previous object
				csg_mat.p_inside_liquid = miFALSE;		//previous object 

				//this need for making standart situation, if shader was called not from p_CSG
				if(cur_first_shd > 0)
				{
					if(cur_object == 1)
					{
						if(!mi_vector_dot(&state->dir, &state->parent->normal_geom))
							csg_mat.inside_glass = miTRUE;
					}
					if(cur_object == 2)
					{
						if(!mi_vector_dot(&state->dir, &state->parent->normal_geom))
							csg_mat.inside_liquid = miTRUE;
					}
				}
			}
//			mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);

			inside_shd * ray_shdStats;
			inside_shd csg_shad;
			ray_shdStats = (inside_shd *)mi_shaderstate_get(state, "csg_shad", NULL);
			if (ray_shdStats != NULL)
			{
				csg_shad = *ray_shdStats;		
				//copy previous value
				csg_shad.p_inside_glass = csg_shad.inside_glass;		
				csg_shad.p_inside_liquid = csg_shad.inside_liquid;
			}
			else //initialise it.
			{
				csg_shad.inside_glass = miFALSE;
				csg_shad.inside_liquid = miFALSE;
				csg_shad.p_inside_glass = miFALSE;
				csg_shad.p_inside_liquid = miFALSE;
			}
			

			if(cur_first_shd > 0)	// first shadow call, we read "csg_mat" (that come from material shader)and set "csg_shad" for read and write for follow shadow calls 
			{
				if(mi_vector_dot(&state->dir, &state->parent->normal_geom) && (csg_mat.inside_glass == miTRUE || csg_mat.inside_liquid == miTRUE))	//we get incorrect "csg_mat" (opposite direction), just restore previous value
				{
					csg_mat.inside_glass = csg_mat.p_inside_glass;
					csg_mat.inside_liquid = csg_mat.p_inside_liquid;
				}

				if((csg_mat.inside_glass == miTRUE) && (csg_mat.inside_liquid == miTRUE))		//insiade A & B
				{
					if(cur_object == 1)
					{
						if(state->inv_normal == miTRUE)
							csg_shad.inside_glass = miFALSE;							
						else		//possible double A
							csg_shad.inside_glass = miTRUE;
						
						mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
						*result = *mi_eval_color(&param->glass_mat);
					//	result->r = result->g = result->b = 0;
					}
					else			//mean object B
					{
						if(cur_object == 2)
						{
							if(state->inv_normal == miTRUE)
								csg_shad.inside_liquid = miFALSE;
							else
								csg_shad.inside_liquid = miTRUE;
							mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
							mi_continue_shadow_seg(result, state);
						}
						else			//error, not defined object name
						{
						//	result->r = result->g = result->b = 0;
							*result = *mi_eval_color(&param->glass_mat);
							return(miTRUE);
						}
					}
				}
				else			//inside A or B or None
				{
					if(csg_mat.inside_glass == miTRUE)				//inside A
					{
						if(cur_object == 1)			//glass
						{
							if(state->inv_normal == miTRUE)
								csg_shad.inside_glass = miFALSE;
							else
								csg_shad.inside_glass = miTRUE;
							mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
							*result = *mi_eval_color(&param->glass_mat);
						}
						else		//mean liquid
						{
							if(state->inv_normal == miTRUE)
								csg_shad.inside_liquid = miFALSE;
							else
								csg_shad.inside_liquid = miTRUE;
							mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
							mi_continue_shadow_seg(result, state);
						}
					}
					else
					{
						if(csg_mat.inside_liquid == miTRUE)				//inside liquid
						{
							if(cur_object == 2)
							{
								if(state->inv_normal == miTRUE)
								{
									csg_shad.inside_liquid = miFALSE;
									mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
									*result = *mi_eval_color(&param->liquid_mat);
								}
								else
								{
									csg_shad.inside_liquid = miTRUE;
									mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
									mi_continue_shadow_seg(result, state);
								}
							}
							else		//mean A
							{
								if(state->inv_normal == miTRUE)
									csg_shad.inside_glass = miFALSE;
								else
									csg_shad.inside_glass = miTRUE;
								mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
								*result = *mi_eval_color(&param->interface_mat);
							}
						}
						else								//inside None
						{
							if(cur_object == 1)
							{									
								if(state->inv_normal == miFALSE)
									csg_shad.inside_glass = miTRUE;
								else
									csg_shad.inside_glass = miFALSE;
								mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
								*result = *mi_eval_color(&param->glass_mat);
							}
							else		//mean B or error
							{
								if(state->inv_normal == miFALSE)
									csg_shad.inside_liquid = miTRUE;
								mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
								*result = *mi_eval_color(&param->liquid_mat);
							}
						}
					}
				}
			}
			else //it secondary shadow call, we use "csg_shad"
			{
				if((csg_shad.inside_glass == miTRUE) && (csg_shad.inside_liquid == miTRUE))		//insiade A & B
				{
					if(cur_object == 1)
					{
						if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
							csg_shad.inside_glass = miFALSE;							
						else		//possible double A
							csg_shad.inside_glass = miTRUE;
						
						mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
						*result = *mi_eval_color(&param->glass_mat);
					}
					else if (cur_object == 2)			//mean object B
					{
						if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
							csg_shad.inside_liquid = miFALSE;
						else
							csg_shad.inside_liquid = miTRUE;
						mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
						mi_continue_shadow_seg(result, state);
					}
					else			//error, not defined object name
					{
						result->r = result->g = result->b = 0;
						return(miTRUE);
					}
				}
				else			//inside A or B or None
				{
					if(csg_shad.inside_glass == miTRUE)				//inside glass
					{
						if(cur_object == 1)
						{
							if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
								csg_shad.inside_glass = miFALSE;
							else
								csg_shad.inside_glass = miTRUE;
							mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
							*result = *mi_eval_color(&param->glass_mat);
						}
						else		//mean liquid
						{
							if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
								csg_shad.inside_liquid = miFALSE;
							else
								csg_shad.inside_liquid = miTRUE;
							mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
							mi_continue_shadow_seg(result, state);
						}
					}
					else if(csg_shad.inside_liquid == miTRUE)				//inside liquid
					{
						if(cur_object == 2)
						{
							if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
								csg_shad.inside_liquid = miFALSE;
							else
								csg_shad.inside_liquid = miTRUE;
							mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
							*result = *mi_eval_color(&param->liquid_mat);
						}
						else		//mean glass
						{
							if(mi_vector_dot(&state->dir, &state->parent->normal_geom))
								csg_shad.inside_glass = miFALSE;
							else
								csg_shad.inside_glass = miTRUE;
							mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
							*result = *mi_eval_color(&param->interface_mat);
						}
					} 
					else								//inside None
					{
						if(cur_object == 1)
						{
							if(!mi_vector_dot(&state->dir, &state->parent->normal_geom))
								csg_shad.inside_glass = miTRUE;
							else
								csg_shad.inside_glass = miFALSE;
							mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
							*result = *mi_eval_color(&param->glass_mat);
						}
						else		//mean B
						{
							if(!mi_vector_dot(&state->dir, &state->parent->normal_geom))
								csg_shad.inside_liquid = miTRUE;
							mi_shaderstate_set(state, "csg_shad", &csg_shad, sizeof(csg_shad), miSS_LIFETIME_EYERAY);
							*result = *mi_eval_color(&param->liquid_mat);
						}
					}					
				}
			}

			if(cur_first_shd == 1)		//cleanup
			{
				mi_shaderstate_set(state, "first_shd", NULL, sizeof(NULL), miSS_LIFETIME_EYERAY);
				mi_shaderstate_set(state, "csg_shad", NULL, sizeof(NULL), miSS_LIFETIME_EYERAY);
			}
		}	
		else       //else segments shadow
		{
			result->r = result->g = result->b = result->a = 0;
			return(miTRUE);
		}
	}
	//////////////////////////////////////////////////////////////////////////////
	else		//else shadow
	{
		inside * rayStats;
		inside csg_mat;
		inside input_csg_mat;		// input state for restore

		rayStats = (inside *)mi_shaderstate_get(state, "csg_mat", NULL);
		if (rayStats != NULL)
		{
			csg_mat = *rayStats;
			input_csg_mat = *rayStats;
			//copy previous value
	//		if(state->type == miRAY_REFLECT || state->type == miRAY_FINALGATHER) //it's reflection ray, just rectore previous value
			if(state->parent != NULL && mi_vector_dot(&state->dir, &state->parent->normal_geom) > 0)
			{
				csg_mat.inside_glass =		csg_mat.p_inside_glass;		
				csg_mat.inside_liquid =		csg_mat.p_inside_liquid;
				csg_mat.p_inside_glass =	csg_mat.p1_inside_glass;		
				csg_mat.p_inside_liquid =	csg_mat.p1_inside_liquid;
				csg_mat.p1_inside_glass =	csg_mat.p2_inside_glass;		
				csg_mat.p1_inside_liquid =	csg_mat.p2_inside_liquid;
				csg_mat.p2_inside_glass =	csg_mat.p3_inside_glass;		
				csg_mat.p2_inside_liquid =	csg_mat.p3_inside_liquid;
			}
			else
			{
				csg_mat.p3_inside_glass =	csg_mat.p2_inside_glass;		
				csg_mat.p3_inside_liquid =	csg_mat.p2_inside_liquid;
				csg_mat.p2_inside_glass =	csg_mat.p1_inside_glass;		
				csg_mat.p2_inside_liquid =	csg_mat.p1_inside_liquid;
				csg_mat.p1_inside_glass =	csg_mat.p_inside_glass;		
				csg_mat.p1_inside_liquid =	csg_mat.p_inside_liquid;
				csg_mat.p_inside_glass =	csg_mat.inside_glass;		
				csg_mat.p_inside_liquid =	csg_mat.inside_liquid;
			}
//			if(state->parent != NULL && mi_vector_dot(&state->dir, &state->parent->normal_geom) > 0)
//			{
//				csg_mat.inside_glass = csg_mat.p_inside_glass;		
//				csg_mat.inside_liquid = csg_mat.p_inside_liquid;
//			}
//			else
//			{
//				csg_mat.p_inside_glass = csg_mat.inside_glass;		
//				csg_mat.p_inside_liquid = csg_mat.inside_liquid;
//			}
		}
		else //initialise it.
		{
			csg_mat.inside_glass = miFALSE;
			csg_mat.inside_liquid = miFALSE;
			csg_mat.p_inside_glass = miFALSE;
			csg_mat.p_inside_liquid = miFALSE;
			csg_mat.p1_inside_glass = miFALSE;
			csg_mat.p1_inside_liquid = miFALSE;
			csg_mat.p2_inside_glass = miFALSE;
			csg_mat.p2_inside_liquid = miFALSE;
			csg_mat.p3_inside_glass = miFALSE;
			csg_mat.p3_inside_liquid = miFALSE;

			input_csg_mat.inside_glass = miFALSE;
			input_csg_mat.inside_liquid = miFALSE;
			input_csg_mat.p_inside_glass = miFALSE;
			input_csg_mat.p_inside_liquid = miFALSE;
			input_csg_mat.p1_inside_glass = miFALSE;
			input_csg_mat.p1_inside_liquid = miFALSE;
			input_csg_mat.p2_inside_glass = miFALSE;
			input_csg_mat.p2_inside_liquid = miFALSE;
			input_csg_mat.p3_inside_glass = miFALSE;
			input_csg_mat.p3_inside_liquid = miFALSE;
		}

		if(*mi_eval_boolean(&param->check_camera) && state->type == miRAY_EYE)
		{
/*			miVector inv_dir = state->dir; 
			mi_vector_neg(&inv_dir);
			if(mi_trace_probe(state, &inv_dir, &state->point))
			{
				if(state->child->instance)
				{
					miInstance *inv_inst = (miInstance *)mi_db_access(state->child->instance);
					const char *inv_obj_name = inv_inst->parent ? mi_api_tag_lookup(inv_inst->parent) : 0;
			//		const char *inv_obj_name = inv_inst->item ? mi_api_tag_lookup(inst->item) : 0;
					mi_db_unpin(state->child->instance);
					
					if(strcmp(inv_obj_name, glass_name) == 0)
					{
						if(state->child->inv_normal)
							csg_mat.inside_glass = miTRUE;
					}
					else if (strcmp(inv_obj_name, liquid_name) == 0)
					{
						if(state->child->inv_normal)
							csg_mat.inside_liquid = miTRUE;
					}
				}
			}
*/
		}

//		int cur_object = current_object(state, param);		//check what object hitted
	

		if((csg_mat.inside_glass == miTRUE) && (csg_mat.inside_liquid == miTRUE))		//insiade A & B
		{
			if(cur_object == 1)
			{
				if(state->inv_normal == miTRUE)
				{
					csg_mat.inside_glass = miFALSE;
					mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
		//			mi_vector_neg(&state->normal);
		//			mi_vector_neg(&state->normal_geom);
					state->inv_normal = miFALSE;
					*result = *mi_eval_color(&param->interface_mat);
				}
				else
				{
					csg_mat.inside_glass = miTRUE;
					mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
					*result = *mi_eval_color(&param->glass_mat);
				}
			}
			else if(cur_object == 2)
			{
				if(state->inv_normal == miTRUE)
					csg_mat.inside_liquid = miFALSE;
				else
					csg_mat.inside_liquid = miTRUE;
				mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
				state->type = miRAY_REFRACT;
				mi_trace_continue(result, state);
			}			
			else			//error
			{
					result->r = 1;
					result->g = result->b = 0;
					return(miTRUE);				
			}
		}
		else			//inside A or B or None
		{
			if(csg_mat.inside_glass == miTRUE)				//inside glass
			{
				if(cur_object == 1)
				{
					if(state->inv_normal == miTRUE)
						csg_mat.inside_glass = miFALSE;
					else
						csg_mat.inside_glass = miTRUE;
					mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
					*result = *mi_eval_color(&param->glass_mat);
				}
				else		//mean liquid
				{
					if(state->inv_normal == miTRUE)
						csg_mat.inside_liquid = miFALSE;
					else
						csg_mat.inside_liquid = miTRUE;
					mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
					state->type = miRAY_REFRACT;
					mi_trace_continue(result, state);
				}
			}
			else
			{
				if(csg_mat.inside_liquid == miTRUE)				//inside liquid
				{
					if(cur_object == 2)		//liquid
					{
						if(state->inv_normal == miTRUE)
							csg_mat.inside_liquid = miFALSE;
						else
							csg_mat.inside_liquid = miTRUE;
						mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
						*result = *mi_eval_color(&param->liquid_mat);
					}
					else		//mean glass
					{
						if(state->inv_normal == miTRUE)
							csg_mat.inside_glass = miFALSE;
						else
							csg_mat.inside_glass = miTRUE;
						mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
	//					mi_vector_neg(&state->normal);
	//					mi_vector_neg(&state->normal_geom);
						state->inv_normal = miTRUE;
						*result = *mi_eval_color(&param->interface_mat);
					}
				}
				else					//inside None
				{
					if(cur_object == 1)
					{
						if(state->inv_normal == miFALSE)
							csg_mat.inside_glass = miTRUE;
						mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
						*result = *mi_eval_color(&param->glass_mat);
					}
					else
					{
						if(state->inv_normal == miFALSE)
							csg_mat.inside_liquid = miTRUE;
						mi_shaderstate_set(state, "csg_mat", &csg_mat, sizeof(csg_mat), miSS_LIFETIME_EYERAY);
						*result = *mi_eval_color(&param->liquid_mat);
					}
				}
			}
		}
	//	restore original input date 
		mi_shaderstate_set(state, "csg_mat", &input_csg_mat, sizeof(input_csg_mat), miSS_LIFETIME_EYERAY);
	//	mi_shaderstate_set(state, "csg_mat", NULL, sizeof(NULL), miSS_LIFETIME_EYERAY);
	}
	return(miTRUE);
}
}
