//  Copyright (c) 2007, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <math.h>
#include "shader.h"

typedef struct
{
	miColor		pupil_color;
	miScalar	pupil_radius;
	miScalar	pupil_catness;
	miScalar	pupil_border_blur;

	miColor		iris_inner_color;
	miScalar	iris_inner_noise_u;
	miScalar	iris_inner_noise_v;
	miScalar	iris_inner_border;
	miScalar	iris_inner_border_noiseness;
	miScalar	iris_inner_border_noise_v;
	miScalar	iris_inner_border_blur;

	miColor		iris_outer_color;
	miScalar	iris_outer_noise_u;
	miScalar	iris_outer_noise_v;
	miScalar	iris_outer_radius;
	miScalar	iris_outer_border_blur;

	miColor		limb_color;
	miScalar	limb_width;
	miScalar	limb_border_blur;

	miColor		sclera_color;
	miColor		sclera_back_color;

	miBoolean	swap_uv;
	miScalar	scale;
}p_eye_t;

static inline miColor p_color_mix (miColor *a, miColor *b, float min, float max, float x)
{
	float smooth;
	miColor result;
	if(min != max)
	{
		float p, term;
		if (x <= min)
			p = min;
		else if (x >= max)
			p = max;
		else
			p = x;
		/* now calculate smoothstep value:
		   -2term^3 + 3term^2, where term = (p-min)/(max-min) */
		term  = (p - min) / (max - min);	//normalize 0 - 1
		smooth = -2 * term * term * term + 3 * term * term;
	}
	else
		smooth = 0.0f;

	result.r = b->r * smooth + a->r *(1-smooth);
	result.g = b->g * smooth + a->g *(1-smooth);
	result.b = b->b * smooth + a->b *(1-smooth);

	return  result;	
}

miScalar p_catness(miScalar radius, miScalar catness, miScalar U, miScalar V)
{
	double cat_scale = U * 2.0f;
	if(cat_scale > 1.0f)
		cat_scale -= 1.0f;
	catness = (1-catness) * radius;
	//get x, y in cartesian coordinates
	double x1 = cos(M_PI_2 *(1-2*cat_scale))*cos(M_PI_2*(1-V));
	double y1 = sin(M_PI_2 *(1-2*cat_scale))*cos(M_PI_2*(1-V));
	double r = (radius*radius + catness*catness)/(2*catness);
	double a = -y1;
	double b = x1;
	double c = y1*(r-catness);
	double x0 = -a*c/(a*a+b*b);
	double y0 = -b*c/(a*a+b*b);
	double d = r*r - c*c/(a*a+b*b);
	double mult = sqrt (d / (a*a+b*b));
	double ax = x0 + b * mult+catness-r;
	double ay = y0 - a * mult;
	return (miScalar)sqrt(ax*ax+ay*ay);
}

extern "C"
{
DLLEXPORT int p_eye_version(void) {return(2);}

DLLEXPORT miBoolean p_eye(miColor *result, miState *state, p_eye_t *paras)
{

	miColor pupil_color =			*mi_eval_color(&paras->pupil_color);
	miScalar pupil_radius =			*mi_eval_scalar(&paras->pupil_radius);
	miScalar pupil_catness =		*mi_eval_scalar(&paras->pupil_catness);
	miScalar pupil_border_blur =	*mi_eval_scalar(&paras->pupil_border_blur);

	miColor iris_inner_color =			*mi_eval_color(&paras->iris_inner_color);
	miScalar iris_inner_noise_u =		*mi_eval_scalar(&paras->iris_inner_noise_u);
	miScalar iris_inner_noise_v =		*mi_eval_scalar(&paras->iris_inner_noise_v);
	miScalar iris_inner_border =		*mi_eval_scalar(&paras->iris_inner_border);
	miScalar iris_inner_border_noiseness = *mi_eval_scalar(&paras->iris_inner_border_noiseness);
	miScalar iris_inner_border_noise_v = *mi_eval_scalar(&paras->iris_inner_border_noise_v);
	miScalar iris_inner_border_blur =	*mi_eval_scalar(&paras->iris_inner_border_blur);

	miColor iris_outer_color =			*mi_eval_color(&paras->iris_outer_color);
	miScalar iris_outer_noise_u =		*mi_eval_scalar(&paras->iris_outer_noise_u);
	miScalar iris_outer_noise_v =		*mi_eval_scalar(&paras->iris_outer_noise_v);
	miScalar iris_outer_radius =		*mi_eval_scalar(&paras->iris_outer_radius);
	miScalar iris_outer_border_blur =	*mi_eval_scalar(&paras->iris_outer_border_blur);

	miColor limb_color =			*mi_eval_color(&paras->limb_color);
	miScalar limb_width =			*mi_eval_scalar(&paras->limb_width);
	miScalar limb_border_blur =		*mi_eval_scalar(&paras->limb_border_blur);
	
	miColor sclera_color =			*mi_eval_color(&paras->sclera_color);
	miColor sclera_back_color =		*mi_eval_color(&paras->sclera_back_color);

	miBoolean swap_uv =			*mi_eval_boolean(&paras->swap_uv);
	miScalar scale =			*mi_eval_scalar(&paras->scale);

	miScalar U;
	miScalar V;
	
//	if(pupil_radius > iris_outer_radius)

	state->tex = *state->tex_list;

	if(swap_uv)		//nurbs
	{
		V = state->tex.x;
		U = state->tex.y;
	}
	else			//poly
	{
		V = state->tex.y;
		U = state->tex.x;
	}
	V = (1-V) * 2.0f;
	if(scale)
		V /= scale;

	if(V < 0)
		V = 0;
	if(V > 1)
		V = 1;
	if(U < 0)
		U = 0;
	if(U > 1)
		U = 1;

	//apply catness
	if(pupil_catness > 0.001f)
		pupil_radius = p_catness(pupil_radius, pupil_catness, U, V);

	iris_inner_border = pupil_radius *(1-iris_inner_border) + iris_outer_radius * iris_inner_border;

	//noise inner border
	if(iris_inner_border_noiseness > 0.001f)
	{
		miScalar iris_inner_border_noise = 0.5f;
		iris_inner_border_noise = (mi_noise_1d(U * iris_inner_border_noise_v)*2.0f -1) * iris_inner_border_noiseness;
		iris_inner_border += iris_inner_border_noise;  //noise radius
	}
	if(iris_inner_border < 0.f)
		iris_inner_border = 0.f;

	miScalar noise_res3 = mi_noise_1d(U*200)*2.0f -1;
	noise_res3 *= 0.1f;

//	U += noise_res3;


	miScalar noise_res1 = mi_noise_2d(U*iris_inner_noise_u, (V-pupil_radius)*iris_inner_noise_v);
	miScalar noise_res2 = mi_noise_2d(U*iris_outer_noise_u, (V-pupil_radius)*iris_outer_noise_v);
	if(noise_res1 < 0)
		noise_res1 = 0.f;
	if(noise_res2 < 0)
		noise_res2 = 0.f;

	iris_inner_color.r *= noise_res1;
	iris_inner_color.g *= noise_res1;
	iris_inner_color.b *= noise_res1;

	iris_outer_color.r *= noise_res2;
	iris_outer_color.g *= noise_res2;
	iris_outer_color.b *= noise_res2;



	if(V <= pupil_radius)				//pupil
		*result = pupil_color;
	else if (V <= iris_outer_radius)
	{
		//mix iris_inner_color and iris_outer_color
		*result = p_color_mix (&iris_inner_color, &iris_outer_color, iris_inner_border, iris_inner_border + iris_inner_border_blur, V);

		if(pupil_border_blur && V <= pupil_radius + pupil_border_blur)
			*result = p_color_mix (&pupil_color, result, pupil_radius, pupil_radius + pupil_border_blur, V);
		if (iris_outer_border_blur && V >= iris_outer_radius - iris_outer_border_blur)
			*result = p_color_mix (result, &limb_color, iris_outer_radius - iris_outer_border_blur, iris_outer_radius, V);
	}
	else if (limb_width && V <= iris_outer_radius + limb_width)						//limb
		*result = limb_color;
	else if (limb_border_blur && V <= iris_outer_radius + limb_width + limb_border_blur)		//limb blur
		*result = p_color_mix (&limb_color, &sclera_color, iris_outer_radius + limb_width, iris_outer_radius + limb_width + limb_border_blur, V);
	else								//sclera
		*result = p_color_mix (&sclera_color, &sclera_back_color, iris_outer_radius + limb_width + limb_border_blur, 1.0f, V);

	result->a = U;
	return(miTRUE);
}
}
