//  Copyright (c) 2005, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "p_fur.h"
#include "p_utils.h"


inline
miScalar fakefurOpacity (const miVector& V, miScalar An, miScalar D, const miVector& T, const miVector& NN)
{
	miScalar g, alpha;

	// we assume that V and T are unit vectors
	//   miVector B = V ^ T;
	miVector B;
	mi_vector_prod(&B, &T, &V);
//	alpha = V % NN;
	alpha = mi_vector_dot(&V, &NN);
//	g = B.length();
	g = p_vector_length(&B);
	if(alpha > miSCALAR_EPSILON)
		g /= alpha;

	// Opacity is an inverted Gaussian
	alpha = exp( D * An * g );
	if (alpha <= miSCALAR_EPSILON)
		return 0.0f;
	alpha = 1 - (1 / alpha);
	return alpha;
}

miVector shiftTangent(miVector T, miVector n, miScalar amt)
{
//   miVector shiftedT = T + n * amt;
	miVector shiftedT;
	miVector nn;
	nn.x = n.x * amt;
	nn.y = n.y * amt;
	nn.z = n.z * amt;
	mi_vector_add(&shiftedT, &T, &nn);

   mi_vector_normalize(&shiftedT);
   return shiftedT;
}

/*
template< typename T >
inline T smoothstep( const T& a, const T& b, const T& x )
{
   if      ( x <= a ) return 0.0f;
   else if ( x >= b ) return 1.0f;
   else
   {
      T t = (x - a) / (b - a);
      return t * t * (3.0f - 2.0f * t);
   }
}
*/

inline miScalar 
smoothstep( const miScalar a, const miScalar b, const miScalar x )
{
   if ( x <= a ) return 0.0f;
   else if ( x >= b ) return 1.0f;
   else
   {
      miScalar result = (x - a) / (b - a);
      return result * result * (3.0f - 2.0f * result);
   }
}

inline miScalar
StrandSpecular(const miVector& T, const miVector& H, const miScalar exponent)
{
//   miScalar TdH = T % H;
	miScalar TdH;
	TdH = mi_vector_dot(&T, &H);
	miScalar sinTH = sqrt( (float)(1.0f - TdH * TdH) );
	miScalar dirAtten = smoothstep(-1.0f, 0.0f, TdH);
	return dirAtten * pow(sinTH, exponent );
}

miColor p_fur_specular(miColor *spec1, miScalar exp1, miColor *spec2, miScalar exp2,
	miVector *hairN, miVector *V, miVector *T, miVector *dir, miColor *light_col)
{
	miColor result;
	result.r = result.g = result.b = 0;

	miVector H;
	mi_vector_add(&H, V, dir);
	mi_vector_normalize(&H);

	miScalar hairNdL;
	hairNdL = mi_vector_dot(hairN, dir);

	hairNdL = StrandSpecular(*T, H, exp1);
	result.r += spec1->r * hairNdL * light_col->r;
	result.g += spec1->g * hairNdL * light_col->g;
	result.b += spec1->b * hairNdL * light_col->b;

	hairNdL = StrandSpecular(*T, H, exp2);
	result.r += spec2->r * hairNdL * light_col->r;
	result.g += spec2->g * hairNdL * light_col->g;
	result.b += spec2->b * hairNdL * light_col->b;

//	if(result.r < 0)
//		result.r = 0;
//	if(result.g < 0)
//		result.g = 0;
//	if(result.b < 0)
//		result.b = 0;

	return result;
}

miColor p_fur_diffuse(miColor *skinColor, miScalar hair_density,
	  miScalar f_backwardScater, miScalar f_forwardScatter,	miScalar f_hairOverHairShad, miScalar f_shadowAtten, miScalar f_startShadowCo,
	  miScalar f_endShadowCo, miScalar hair_area, miVector *hairN, miVector *V, miVector *T, miVector *dir, miColor *light_col, miScalar dot_nl)
{
   miScalar f_dir;	// directional attenuation factor
   miScalar f_surface;	// surface normal factor
   miScalar hair_over_skin_shadow;
   miScalar hair_over_hair;
   miScalar hair_skin_visibility;
   miColor alpha_skin;
   miColor alpha_hair;
   miColor kajiya;

	miVector TcL;
	miScalar hairNdL;

	miColor result;
	result.r = result.g = result.b = 0;

			hair_over_hair = 1 - f_hairOverHairShad * fakefurOpacity(*dir, hair_area, hair_density, *T, *hairN );

			miVector H;
			mi_vector_add(&H, V, dir);
			mi_vector_normalize(&H);

			hairNdL = mi_vector_dot(hairN, dir);
			kajiya.r = light_col->r * hairNdL;
			kajiya.g = light_col->g * hairNdL;
			kajiya.b = light_col->b * hairNdL;

			mi_vector_prod(&TcL, T, dir);

			f_dir = f_backwardScater + f_forwardScatter;

			f_surface = 1 + f_shadowAtten * ( smoothstep( f_startShadowCo, f_endShadowCo, mi_vector_dot(hairN, dir)) - 1);

			alpha_hair.r = f_dir * f_surface * kajiya.r;
			alpha_hair.g = f_dir * f_surface * kajiya.g;
			alpha_hair.b = f_dir * f_surface * kajiya.b;

			alpha_hair.r *= hair_over_hair;
			alpha_hair.g *= hair_over_hair;
			alpha_hair.b *= hair_over_hair;

			hair_over_skin_shadow = 1 - fakefurOpacity(*dir, hair_area, hair_density, *T, *hairN );

			alpha_skin.r = skinColor->r * dot_nl * light_col->r;
			alpha_skin.g = skinColor->g * dot_nl * light_col->g;
			alpha_skin.b = skinColor->b * dot_nl * light_col->b;

			alpha_skin.r *= hair_over_skin_shadow;
			alpha_skin.g *= hair_over_skin_shadow;
			alpha_skin.b *= hair_over_skin_shadow;

			hair_skin_visibility = fakefurOpacity(*V, hair_area, hair_density, *T, *hairN );

			result.r += ( hair_skin_visibility * alpha_hair.r + ( 1 - hair_skin_visibility ) * alpha_skin.r );
			result.g += ( hair_skin_visibility * alpha_hair.g + ( 1 - hair_skin_visibility ) * alpha_skin.g );
			result.b += ( hair_skin_visibility * alpha_hair.b + ( 1 - hair_skin_visibility ) * alpha_skin.b );

//			result.r = hair_skin_visibility;
//			result.g = hair_skin_visibility;
//			result.b = hair_skin_visibility;
			return result;
}

inline
miColor p_fur(miColor *skinColor, miColor *hairColor, miColor *spec1, miScalar exp1, miColor *spec2, miScalar exp2, miScalar hair_density,
	  miScalar f_backwardScater, miScalar f_forwardScatter,	miScalar f_hairOverHairShad, miScalar f_shadowAtten, miScalar f_startShadowCo,
	  miScalar f_endShadowCo, miScalar bup, miScalar hair_area, miVector *hairN, miVector *V, miVector *T, miVector *dir, miColor *light_col, miScalar dot_nl)
{

//   miScalar k;	        // relative direction of a given incident light ray
   miScalar f_dir;	// directional attenuation factor
   miScalar f_surface;	// surface normal factor
   miScalar hair_over_skin_shadow;
   miScalar hair_over_hair;
   miScalar hair_skin_visibility;
   miColor alpha_skin;
   miColor alpha_hair;
   miColor kajiya;

	// Cross product of the tangent along the hair
	// and light direction
	miVector TcL;
	miScalar hairNdL;

	miColor result;
	result.r = result.g = result.b = 0;

			// using the fakefur opacity function, compute the hairover-hair
			// shadow attenuation
//			hair_over_hair = 1 - s * fakefurOpacity( L, hair_area, hair_density, T, hairN );
			hair_over_hair = 1 - f_hairOverHairShad * fakefurOpacity(*dir, hair_area, hair_density, *T, *hairN );

//			miVector H = V + L;
			miVector H;
			mi_vector_add(&H, V, dir);
			mi_vector_normalize(&H);

			// Kajiya contribution
//			hairNdL = hairN % L;
			hairNdL = mi_vector_dot(hairN, dir);
//			kajiya = diffuse * Cl * hairNdL;
			kajiya.r = hairColor->r * light_col->r * hairNdL;
			kajiya.g = hairColor->g * light_col->g * hairNdL;
			kajiya.b = hairColor->b * light_col->b * hairNdL;
//			kajiya.r = kajiya.g = kajiya.b = 0;

			hairNdL = StrandSpecular(*T, H, exp1);
//			kajiya += spec1 * hairNdL * Cl;
			kajiya.r += spec1->r * hairNdL * light_col->r;
			kajiya.g += spec1->g * hairNdL * light_col->g;
			kajiya.b += spec1->b * hairNdL * light_col->b;

			hairNdL = StrandSpecular(*T, H, exp2);
//			kajiya += spec2 * hairNdL * Cl;
			kajiya.r += spec2->r * hairNdL * light_col->r;
			kajiya.g += spec2->g * hairNdL * light_col->g;
			kajiya.b += spec2->b * hairNdL * light_col->b;

			// scattering of light by fur
//			TcL = T ^ L;
			mi_vector_prod(&TcL, T, dir);

			// directional attenuation factor
//			f_dir = ((1 + k) / 2 * p_reflect) + ((1 - k) / 2 * p_transmit);
			f_dir = f_backwardScater + f_forwardScatter;

			// Surface normal factor as a quick and dirty way to adjust shadowing
//			f_surface = 1 + p_surface * ( smoothstep( w_min, w_max, hairN % L ) - 1 );
			f_surface = 1 + f_shadowAtten * ( smoothstep( f_startShadowCo, f_endShadowCo, mi_vector_dot(hairN, dir)) - 1);

			// compute the reflected luminance of the average hair
			// in the sample region
//			alpha_hair = f_dir * f_surface * kajiya;
			alpha_hair.r = f_dir * f_surface * kajiya.r;
			alpha_hair.g = f_dir * f_surface * kajiya.g;
			alpha_hair.b = f_dir * f_surface * kajiya.b;
//			alpha_hair *= hair_over_hair;
			alpha_hair.r *= hair_over_hair;
			alpha_hair.g *= hair_over_hair;
			alpha_hair.b *= hair_over_hair;

			// using the fakefur opacity function, compute the
			// hair-over-skin shadow factor
//			hair_over_skin_shadow = 1 - fakefurOpacity( L, hair_area, hair_density, T, hairN );
			hair_over_skin_shadow = 1 - fakefurOpacity(*dir, hair_area, hair_density, *T, *hairN );

			// compute the reflected luminance of the underlying skin
//			NdL = skinN % L;
//			dot_nl = mi_vector_dot(&skinN, &dir);
//			alpha_skin  = skinColor * NdL * Cl;
			alpha_skin.r = skinColor->r * dot_nl * light_col->r;
			alpha_skin.g = skinColor->g * dot_nl * light_col->g;
			alpha_skin.b = skinColor->b * dot_nl * light_col->b;

			alpha_skin.r *= hair_over_skin_shadow;
			alpha_skin.g *= hair_over_skin_shadow;
			alpha_skin.b *= hair_over_skin_shadow;

			// compute the hair/skin visibility ratio
//			hair_skin_visibility = fakefurOpacity( V, hair_area, hair_density, T, hairN );
			hair_skin_visibility = fakefurOpacity(*V, hair_area, hair_density, *T, *hairN );

			// blend the reflected luminance of the skin and hair using the
			// visibility ratio to obtain the final reflected luminance of
			// the sample region.
//			Cs += ( hair_skin_visibility * alpha_hair + ( 1 - hair_skin_visibility ) * alpha_skin );
			result.r += ( hair_skin_visibility * alpha_hair.r + ( 1 - hair_skin_visibility ) * alpha_skin.r );
			result.g += ( hair_skin_visibility * alpha_hair.g + ( 1 - hair_skin_visibility ) * alpha_skin.g );
			result.b += ( hair_skin_visibility * alpha_hair.b + ( 1 - hair_skin_visibility ) * alpha_skin.b );

//			result.r = hair_skin_visibility;
//			result.g = hair_skin_visibility;
//			result.b = hair_skin_visibility;

//			result.r = alpha_hair.r;
//			result.g = alpha_hair.g;
//			result.b = alpha_hair.b;
//			return alpha_skin;

			return result;
}
