#include "p_occlusion.h"
#include "p_utils.h"


inline miScalar linear(const miScalar &a, const miScalar &b, const miScalar &x)
{
	if(a != b)
	{
		if      ( x <= a ) return 0.0f;
		else if ( x >= b ) return 1.0f;
		else
			return (x - a) / (b - a);
	}
	else
		return a;
}

inline miUint mix(const miUint a, const miUint b, const miScalar x)
{
   if ( x <= 0.0f ) return a;
   if ( x >= 1.0f ) return b;
   return (miUint)(a + (b - a) * x);
}


miBoolean p_occlusion(miColor *result, miVector *bent_result, miState *state, struct static_param *sp, miColor *base_color, miColor *dirt_color, miVector basis, miScalar blend)
{
	miGeoScalar	nearClipS;
	miGeoScalar	farClipS;
	miScalar	hitDist;		// distance to the first hit
	miColor		hitSum;			// averaged sum of all hits
	miColor		curHit;			// a little helper for color computations
	miColor		curEnvironment; // contains the currently sampled environment
	double		sample[2];		// 2d sample, will be filled by mi_sample
	miInteger	sampleCnt;		// sample counter, needed for mi_sample
	miVector	sampleDir, n;	// unit direction vector ray will be shot at, and the current normal n
	
	if(blend > 0.f)	//we use basis vector, suppose basis vector is in world space
	{
		miVector point = state->point;

		mi_point_to_world(state, &point, &point);
		mi_vector_sub(&basis, &basis, &point);
		mi_normal_from_world(state, &basis, &basis);

		mi_vector_normalize(&basis);
	}

	miOptions *orig_options = (miOptions *)(state->options);		//save pointer to original options
	miOptions m_options;
	if(sp->o_opacity_mode == 1)										//use shadow shader for occlusion trasparency
	{
		//define occlusion ray type for p_MegaTK shader
		int curOcc = 1;
		mi_shaderstate_set(state, "ra_o", &curOcc, sizeof(curOcc), miSS_LIFETIME_EYERAY);
		//we don't need following features for computing trasparency, just disable it temporary
		m_options.reflection_depth = 0;								//disable reflection
		m_options.refraction_depth = 0;								//disable refraction, maybe we need it?
		m_options.finalgather = 0;									//disable final gather
		m_options.globillum = miFALSE;								//disable GI
		m_options.shadow = 0;										//disable shadow

	}
	else if(sp->o_opacity_mode == 2)								//use material shader for occlusion trasparency
	{
		m_options = *state->options;								//copy original options
		state->options = &m_options;								//link new options to state
		//edit new options
		m_options.shadow = 1;										//set shadow type to sort		
	}
	

	// eval those two first - if they are 0, no work has to be done
	miUint rayCount = sp->o_samples;

	// if this is zero, we get to do nothing
	if (rayCount <= 0)
	{
		result->r = result->g = result->b = 1.0f;
		return miTRUE;
	}

	if ((sp->o_inv_sec_illum == miTRUE) && (state->type == miRAY_FINALGATHER))
	{
		result->r = result->g = result->b = 1.0f;
		return miTRUE;
	}

	// seems like all the early checking didn't save us from doing some
	// real work - eval all (most) other inputs

	nearClipS = sp->o_near_clip;
	farClipS = sp->o_far_clip;

	sampleCnt = 0;

	hitSum.r = hitSum.g = hitSum.b = 0;

	mi_ray_falloff(state, &nearClipS, &farClipS);

	if(sp->o_falloff_gain != 0)
		sp->o_falloff_gain = 1.0f / sp->o_falloff_gain;

	// we can precompute that, it never changes in the loop
	n = state->normal;

	// if blend = 0, we always use the normal, so skip this:
	if(blend > 0.0f)
	{
//		rayDir = *mi_eval_vector(&paras->dir);
		// multiply the normal by 1 - blend
		mi_vector_mul(&n, 1.0f - blend);

		// and multiply it with blend
		mi_vector_mul(&basis, blend);

		// then add the two together
		mi_vector_add(&n, &basis, &n);

		// and normalize again, so we get a new unit direction vector
		mi_vector_normalize(&n);
	}

	// now, pre-multiply the direction vector with spread (this also
	// never changes in the loop)
//	mi_vector_mul(&n, 1.0f - (sp->o_spread / 180.0f));
	mi_vector_mul(&n, 1.0f - (sp->o_spread * 0.00555f));

	if(sp->o_trace_environment)
	{
		// get a 2-d sample from mi's QMC sampler
		while(mi_sample(sample, &sampleCnt, state, 2, &rayCount))
		{
			// map the sample to a sphere
			sampleDir = p_sphere_sample(sample);

			// premultiply this
//			mi_vector_mul(&sampleDir, sp->o_spread / 180.0f);
			mi_vector_mul(&sampleDir, sp->o_spread * 0.00555f);
			
			// add it together
			mi_vector_add(&sampleDir, &n, &sampleDir);

			// sampleDir now holds our sampling direction, so lets rock
			// first, trace an environment ray
//				mi_trace_environment(&curEnvironment, state, &sampleDir);
//				curEnvironment.r = curEnvironment.g = curEnvironment.b = 1;

				// then, send a probe into the scene, but ONLY if probeGeom
				// is true!!! thanks to lazy eval, we can pack that into
				// a single if
			if(mi_trace_probe(state, &sampleDir, &(state->point)))
			{
				switch(sp->o_opacity_mode)
				{
					case 0:				//opaque
						hitDist = (miScalar)(state->child->dist);
						break;
					case 1:				//material shader
						miColor mat_result;
						mat_result.r = mat_result.g = mat_result.b = mat_result.a = 1;

						miState grandchild;
						state->child->child = &grandchild;
						miVector tex_list_buffer[MAX_TEX];
						state->child->tex_list = tex_list_buffer;
	   
						// Fill out state->tex_list
						for(int j=0; ; j++) 
						{
							const miVector *ta, *tb, *tc;
							if (!mi_tri_vectors(state->child, 't', j, &ta, &tb, &tc))
								break;
							mib_lightmap_combine_vectors(&state->child->tex_list[j], ta, tb, tc, (miVector *)state->child->bary); 
						}

						hitDist = (miScalar)(state->child->dist);
						
	//					mat_nearClipS = static_nearClipS;
	//					mat_farClipS = static_farClipS - state->child->dist;  //minus ray dist

	//					mi_ray_falloff(state->child, &mat_nearClipS, &mat_farClipS);

						if(mi_call_material(&mat_result, state->child))
						{
							if(mat_result.a < 0.3)
									hitDist = sp->o_far_clip;
							else
								hitDist = (miScalar)(state->child->dist);
						}
	//					mat_nearClipS = static_nearClipS;
	//					mat_farClipS = static_farClipS;
	//					mi_ray_falloff(state->child, &mat_nearClipS, &mat_farClipS);
//						else
//							mi_warning("Material Shader Not Called");
						break;
					case 2:				//shadow shader
						miColor shd_result;
						shd_result.r = shd_result.g = shd_result.b = shd_result.a = 1.0f;

						miState grandchild_shd;
						state->child->child = &grandchild_shd;
						miVector tex_list_buffer_shd[MAX_TEX];
						state->child->tex_list = tex_list_buffer_shd;
	   
						/* Fill out state->tex_list */
						for(int j=0; ; j++) 
						{
							const miVector *ta, *tb, *tc;
							if (!mi_tri_vectors(state->child, 't', j, &ta, &tb, &tc))
								break;
							mib_lightmap_combine_vectors(&state->child->tex_list[j], ta, tb, tc, (miVector *)state->child->bary); 
						}

						hitDist = (miScalar)(state->child->dist);
						state->child->type = miRAY_SHADOW;			//ray type should be shadow for shadow shaders
						if(mi_call_shadow_material(&shd_result, state->child))
						{
							if(((shd_result.r + shd_result.g + shd_result.b)*0.3333f) > 0.7)
								hitDist = sp->o_far_clip;
							else
								hitDist = (miScalar)(state->child->dist);
						}
//						else
//							mi_warning("Shadow Shader Not Called");
						break;
					default:			//opaque
						hitDist = (miScalar)(state->child->dist);
						break;
				}

				// are we below the near threshold?
				if (hitDist <= sp->o_near_clip)
				{
					// apply full dirt
//					curEnvironment.r = curEnvironment.g = curEnvironment.b = 1;
//					p_color_mul(&curHit, &dirtColor, &curEnvironment);
					curHit = *dirt_color;
					p_color_add(&hitSum, &hitSum, &curHit);
				}
				else
				{
					mi_trace_environment(&curEnvironment, state, &sampleDir);
					if (hitDist >= sp->o_far_clip)
					{
						p_color_mul(&curHit, base_color, &curEnvironment);
						p_color_add(&hitSum, &hitSum, &curHit);
					}
					else
					{
						// we are somewhere inbetween, so get a falloff from 0 - 1
						// also apply the gain, by using pow(falloff, 1 - gain)
						p_color_mix(&curHit, base_color, dirt_color, (miScalar)pow(1.0f - ((hitDist - sp->o_near_clip) / (sp->o_far_clip - sp->o_near_clip)), sp->o_falloff_gain));
						p_color_mul(&curHit, &curHit, &curEnvironment);
						p_color_add(&hitSum, &hitSum, &curHit);
					}
				}
			}
			else	// we didn't hit anything, so apply full base
			{
				mi_trace_environment(&curEnvironment, state, &sampleDir);
				p_color_mul(&curHit, base_color, &curEnvironment);
				p_color_add(&hitSum, &hitSum, &curHit);

				if(sp->o_bent_normal)
				{
					bent_result->x += sampleDir.x;
					bent_result->y += sampleDir.y;
					bent_result->z += sampleDir.z;
				}
			}
		}
	}
	else // ELSE (multEnviro)
	{
		// this is the same again as above, but without respecting the environment
		while(mi_sample(sample, &sampleCnt, state, 2, &rayCount))
		{
			sampleDir = p_sphere_sample(sample);
//			mi_vector_mul(&sampleDir, sp->o_spread / 180.0f);
			mi_vector_mul(&sampleDir, sp->o_spread * 0.00555f);
			mi_vector_add(&sampleDir, &n, &sampleDir);

			// sampleDir now holds our sampling direction, so lets rock

			// while checking for probeGeom here might seem really stupid
			// (if it is true, we return full base color - always!), we
			// never know what the customer wants to do with our nice little
			// shader.
			if (mi_trace_probe(state, &sampleDir, &(state->point)))
			{
				switch(sp->o_opacity_mode)
				{
					case 0:				//opaque
						hitDist = (miScalar)(state->child->dist);
						break;
					case 1:				//material shader
						miColor mat_result;
						mat_result.r = mat_result.g = mat_result.b = mat_result.a = 1.0f;

						miState grandchild;
						state->child->child = &grandchild;
						miVector tex_list_buffer[MAX_TEX];
						state->child->tex_list = tex_list_buffer;
	   
						/* Fill out state->tex_list */
						for(int j=0; ; j++) 
						{
							const miVector *ta, *tb, *tc;
							if (!mi_tri_vectors(state->child, 't', j, &ta, &tb, &tc))
								break;
							mib_lightmap_combine_vectors(&state->child->tex_list[j], ta, tb, tc, (miVector *)state->child->bary); 
						}

						hitDist = (miScalar)(state->child->dist);
						if(mi_call_material(&mat_result, state->child))
						{
							if(mat_result.a < 0.3)
									hitDist = sp->o_far_clip;
							else
								hitDist = (miScalar)(state->child->dist);
						}
//						else
//							mi_warning("Material Shader Not Called");
						break;
					case 2:				//shadow shader
						miColor shd_result;
						shd_result.r = shd_result.g = shd_result.b = shd_result.a = 1.0f;

						miState grandchild_shd;
						state->child->child = &grandchild_shd;
						miVector tex_list_buffer_shd[MAX_TEX];
						state->child->tex_list = tex_list_buffer_shd;
	   
						/* Fill out state->tex_list */
						for(int j=0; ; j++) 
						{
							const miVector *ta, *tb, *tc;
							if (!mi_tri_vectors(state->child, 't', j, &ta, &tb, &tc))
								break;
							mib_lightmap_combine_vectors(&state->child->tex_list[j], ta, tb, tc, (miVector *)state->child->bary); 
						}

						hitDist = (miScalar)(state->child->dist);
						state->child->type = miRAY_SHADOW;			//ray type should be shadow for shadow shaders
						if(mi_call_shadow_material(&shd_result, state->child))
						{
//							float eee = (shd_result.r + shd_result.g + shd_result.b)/3;
//							mi_warning("shadow result = %f", eee);
							if(((shd_result.r + shd_result.g + shd_result.b)*0.3333f) > 0.7)
								hitDist = sp->o_far_clip;
							else
								hitDist = (miScalar)(state->child->dist);
						}
//						else
//							mi_warning("Shadow Shader Not Called");
						break;
					default:			//opaque
						hitDist = (miScalar)(state->child->dist);
						break;
				}

				if (hitDist <= sp->o_near_clip)
				{
					p_color_add(&hitSum, &hitSum, dirt_color);
					curHit = *dirt_color;
				}
				else if (hitDist >= sp->o_far_clip)
				{
					p_color_add(&hitSum, &hitSum, base_color);
					curHit = *base_color;
				}
				else
				{
					p_color_mix(&curHit, base_color, dirt_color, (miScalar)pow(1.0f - ((hitDist - sp->o_near_clip) / (sp->o_far_clip - sp->o_near_clip)), sp->o_falloff_gain));
					p_color_add(&hitSum, &hitSum, &curHit);
				}
			}
			else
			{
				p_color_add(&hitSum, &hitSum, base_color);
//				mi_vector_add(&bent_result, &bent_result, &sampleDir);
				if(sp->o_bent_normal)
				{
					bent_result->x += sampleDir.x;
					bent_result->y += sampleDir.y;
					bent_result->z += sampleDir.z;
				}
			}
		}
	} //ENDIF (multEnviro)

	// now, divide hitSum through the number of samples, so we get an average
	result->r = hitSum.r / sampleCnt;
	result->g = hitSum.g / sampleCnt;
	result->b = hitSum.b / sampleCnt;
	result->a = hitSum.a / sampleCnt;

	//gamma
	if(sp->o_gamma != 1)
	{
		if(result->r != 1 || result->r != 0)
			result->r = pow(result->r, sp->o_gamma);
		if(result->g != 1 || result->g != 0)
			result->g = pow(result->g, sp->o_gamma);
		if(result->b != 1 || result->b != 0)
			result->b = pow(result->b, sp->o_gamma);
	}

	// restore mi_ray_fallof to the original values
	mi_ray_falloff(state, &nearClipS, &farClipS);
	state->options = orig_options;				//restore pointer to original options
	mi_vector_normalize(bent_result);

	switch(sp->o_bent_space)
	{
		case 0:			//world / internal
			break;
		case 1:			//object
			mi_vector_to_object(state, bent_result, bent_result);
			break;
		case 2:			//camera
			mi_vector_to_camera(state, bent_result, bent_result);
			break;
//		case 3:			//light
//			mi_vector_to_light(state, &bent_result, &bent_result);
//			break;
		default:		//world / internal
			break;
	}
	if(sp->o_opacity_mode == 1)
		mi_shaderstate_set(state, "ra_o", NULL, sizeof(NULL), miSS_LIFETIME_EYERAY);

//	result->r = bent_result.x;
//	result->g = bent_result.y;
//	result->b = bent_result.z;
	return miTRUE;
}
