#include "p_reflect_refract.h"
#include "p_deriv.h"
#include "p_utils.h"


typedef struct
{
	miBoolean first;
	miBoolean second;
	miBoolean third;
}ray_type;

/* Test if we are already multisampling */

typedef struct mi_glossy_state_t 
{
    int multisampling;
} mi_glossy_state;

static miBoolean test_multisampling(miState *state)
{
    int keyz = 4;
    mi_glossy_state *ss = (mi_glossy_state*)mi_shaderstate_get(state, "miMS", &keyz);
    if (!ss) return miFALSE;
    return (miBoolean)ss->multisampling;
}

static void set_multisampling(miState *state, miBoolean mode)
{
    int keyz = 4;
    mi_glossy_state *ss = (mi_glossy_state*)mi_shaderstate_get(state, "miMS", &keyz);

    if (!ss)
    {
        mi_glossy_state ns;
        ns.multisampling = mode;
        mi_shaderstate_set(state, "miMS", &ns, sizeof(ns), miSS_LIFETIME_EYERAY);
    }
    else
        ss->multisampling = mode;
}

inline void p_anis_orientation(miVector *u, miVector *v, miState *state)
{
	float d;
	*u = state->derivs[0];
	d  = mi_vector_dot(u, &state->normal);
	u->x -= d * state->normal.x;
	u->y -= d * state->normal.y;
	u->z -= d * state->normal.z;
	miASSERT(mi_vector_dot(u, &state->normal) < miEPS);
	mi_vector_normalize(u);

	// Set v to be perpendicular to u (in the tangent plane) 
	mi_vector_prod(v, &state->normal, u);
	miASSERT(fabs(mi_vector_norm(u) - 1) < miEPS);
	miASSERT(fabs(mi_vector_norm(v) - 1) < miEPS);
	miASSERT(fabs(mi_vector_dot(u, v)) < miEPS);
}

miCBoolean locate_volume(miState *state, miScalar &ior_in, miScalar &ior_out, miBoolean ignore_normals)
{
   miState     *s, *s_in = NULL;
   miCBoolean  enter_surf;
   int         num = 0;

   enter_surf = !state->inv_normal;
   for (s=state; s; s=s->parent)
      if ((s->type == miRAY_TRANSPARENT || 
		  s->type == miRAY_REFRACT || 
		  s->type == miPHOTON_TRANSMIT_SPECULAR || 
		  s->type == miPHOTON_TRANSMIT_GLOSSY || 
		  s->type == miPHOTON_TRANSPARENT || 
		  s->type == miPHOTON_TRANSMIT_DIFFUSE) && 
		  s->parent && s->parent->shader == state->shader)
	  {
         num++;
         if (!s_in)
			s_in = s->parent;
      }
	if (ignore_normals)
		enter_surf = !(num & 1);

	if (enter_surf)		// entering
	{   
		if(!state->refraction_volume)
			state->refraction_volume = state->volume;
//		if (is_interface)
			ior_out = state->parent && state->parent->ior != 0 ? state->parent->ior : 1;
		state->ior_in = ior_out;
		state->ior = ior_in;
	} 
	else 
	{
		state->refraction_volume = s_in ? s_in->volume : state->camera->volume;
//		if (is_interface)
			ior_out = s_in && s_in->ior_in != 0 ? s_in->ior_in : 1;
		state->ior = ior_out;
		state->ior_in = ior_in;
	}
	return(enter_surf);
}

/* Calculate the u/v deviation vectors */

static void calculate_vectors(miState *state, miVector *u_out, 
                              miVector *v_out, miVector *u_in, 
                              miVector *v_in, miBoolean anisotropic)
{
    miBoolean have_v = miFALSE;

    if (anisotropic)
    {
        /* Eval the param */
        u_in = mi_eval_vector(u_in);

        /* If the user specified nothing */
        if (u_in->x == 0.0 && u_in->y == 0.0 && u_in->z == 0)
        {
            /* Try a derivative, if one exists */
            *u_out = state->derivs[0];

            /* If it was a null vector */
            if (u_out->x == 0.0 && u_out->y == 0.0 && u_out->z == 0.0)
            {
                miVector  foo;      /* Arb. vector */
                int loop = 2;		/* Loop count  */
                miScalar  dot_prod; /* Dot product */
                /* Pick an arbitrary vector, object space Z-axis */
                foo.x = 0.0;
                foo.y = 0.0;
                foo.z = 1.0;

                /* Loop exists to handle degenerate case 
                   normally executed once only */
                while (loop-- > 0) 
				{
                    /* Convert from object space */
                    mi_vector_from_object(state, v_out, &foo);

                    dot_prod = mi_vector_dot(&state->normal, v_out);

                    /* Force vector to be perpendicular */
                    v_out->x -= state->normal.x * dot_prod;
                    v_out->y -= state->normal.y * dot_prod;
                    v_out->z -= state->normal.z * dot_prod;

                    /* Low quality case */
                    if (dot_prod > 0.999)
                    {
                        /* Pick different vector, Y */
                        foo.x = 0.0;
                        foo.y = 1.0;
                        foo.z = 0.0;
                        continue; /* Re-loop */
                    }
                    break;                   
                }
                /* We calculated v, so no need to re-do that */
                have_v = miTRUE;
                /* Create u_out perp to v_out and normal */
                mi_vector_prod(u_out, &state->normal, v_out); 
            }
        }
        else /* Use users vector */
        {
            mi_vector_from_world(state, u_out, u_in);
           
            /* Eval the param */
            v_in = mi_eval_vector(v_in);

            /* If the user specified something */
            if (v_in->x != 0.0 || v_in->y != 0.0 || v_in->z == 0)
            {
                mi_vector_from_world(state, v_out, v_in);
                mi_vector_normalize(v_out);
                have_v = miTRUE;
            }
        }

        mi_vector_normalize(u_out);

        if (!have_v)
        {
            /* Create v_out perp to u_out and normal */
            mi_vector_prod(v_out, &state->normal, u_out);
        }
    }
    else /* Isotropic, direction doesnt matter */
    {
        /* Build arbitrary perp vectors by transposing 
           x, y and z components */
        u_out->x = -state->normal.z;
        u_out->y = state->normal.x;
        u_out->z = state->normal.y;

        /* Create v_out perp to u_out and normal */
        mi_vector_prod(v_out, &state->normal, u_out);
        /* Create u_out perp to v_out and normal */
        mi_vector_prod(u_out, &state->normal, v_out);    
    }
}

miColor rainbowdef[] = 
{
    { 1, 0, 0} ,  /* Red */
    { 1, 1, 0} ,  /* Yellow */
    { 1, 1, 1} ,  /* White */
    { .5, 1, 1} , /* Cyan-ish */
    { 0, 0, 1} ,  /* Blue */
    { 1, 0, 1}    /* Indigo (Magenta) */
};

static void affect_rainbow(miColor *color, miScalar colorfac, miScalar strength, miColor* rainbow, int n_items)
{
    int idx, idx2;
    miScalar fac1, fac2;
    miColor color1, color2, color3;

    /* We want to adress UP TO n_items, so subtract one, i.e. for
       6 array items we want to go from 0 to 5, not 6 */
    n_items--;

    /* change +/- 0.5 range to a 0.0-1.0 range) */
    colorfac += 0.5;

    /* Assure 0 to 1 range */
    if (colorfac < 0) colorfac = 0;
    if (colorfac > 1) colorfac = 1;

    colorfac *= (float)n_items; 

    idx  = (int)floor(colorfac);
    idx2 = idx+1;

    if (idx >  n_items) idx = n_items;
    if (idx2 > n_items) idx2 = n_items;

    fac2 = colorfac - idx;
    fac1 = 1.0f - fac2;

    color1 = rainbowdef[idx];
    color2 = rainbowdef[idx2];

    color3.r = ((color1.r * fac1 + color2.r * fac2) * strength) 
               + (1.0f - strength);
    color3.g = ((color1.g * fac1 + color2.g * fac2) * strength) 
               + (1.0f - strength);
    color3.b = ((color1.b * fac1 + color2.b * fac2) * strength) 
               + (1.0f - strength);

    color->r *= color3.r;
    color->g *= color3.g;
    color->b *= color3.b;
}

miBoolean p_reflect(miColor *result, miState *state, p_MegaTK_t *paras, miColor *glossy, miScalar *shiny,
                    miBoolean anisotropic, miScalar *shiny_u, miScalar *shiny_v,
                    miScalar falloff_start, miScalar falloff_stop, 
					miBoolean falloff_to_color, miColor *falloff_color, int samples, int samples_decrease)
{
	if(samples < 1)
		return(miTRUE);

	miVector rdir;
	miColor  rcol;
	rdir.x = rdir.y = rdir.z = 0;
	rcol.r = rcol.g = rcol.b = rcol.a = 0;
//	int      n_spectrum   = 0;
//	miColor *spectrum     =  NULL;    
//	miScalar  rainbow = *mi_eval_scalar(&paras->rainbowness_refl);

	if(falloff_start > falloff_stop)
		falloff_start = falloff_stop;
	miBoolean falloff_changed = miFALSE;
	miGeoScalar falloff_near = falloff_start;
	miGeoScalar falloff_far = falloff_stop;

	if(falloff_stop > 0.f && !falloff_to_color)	//limit ray lenght
	{
		mi_ray_falloff(state, &falloff_near, &falloff_far);
		falloff_changed = miTRUE;
	}

//	if (rainbow > 0.f)
//    {
//        n_spectrum = *mi_eval_integer(&paras->n_spectrum_refl);
//        spectrum = paras->spectrum_refl + *mi_eval_integer(&paras->i_spectrum_refl);
//        if (n_spectrum == 0)
//        {
//            spectrum = rainbowdef;
//            n_spectrum = 6;
//        }            
//    }

	if(!*shiny && !anisotropic)		//specular reflect
	{
		mi_reflection_dir_specular(&rdir, state);
		if(mi_trace_reflection(&rcol, state, &rdir) || mi_trace_environment(&rcol, state, &rdir))
		{
			double dist = 0.f;
			if (state->child) // zero distance for environment calls
				dist = state->child->dist;
			if(falloff_stop > 0.f && dist && falloff_to_color) //falloff to color used
			{
				if(falloff_start < dist && falloff_stop > dist)
				{
//					float atten = 1.f;
//					atten = (float)exp(-(dist -falloff_start) / (falloff_stop - falloff_start));
					float atten = (float)(dist -falloff_start) / (falloff_stop - falloff_start);
					result->r = (rcol.r * (1-atten) + falloff_color->r * atten) * glossy->r;
					result->g = (rcol.g * (1-atten) + falloff_color->g * atten) * glossy->g;
					result->b = (rcol.b * (1-atten) + falloff_color->b * atten) * glossy->b;
				}
				else if (dist >= falloff_stop)		//full falloff color
				{
					result->r = falloff_color->r * glossy->r;
					result->g = falloff_color->g * glossy->g;
					result->b = falloff_color->b * glossy->b;
				}
				else if (dist <= falloff_start)		//no falloff color
				{
					result->r = rcol.r * glossy->r;
					result->g = rcol.g * glossy->g;
					result->b = rcol.b * glossy->b;
				}
			}
			else if (falloff_stop > 0.f && falloff_to_color)		//environment call with falloff color
			{
				result->r = (rcol.r * (1-falloff_color->a) + falloff_color->r * falloff_color->a) * glossy->r;
				result->g =	(rcol.g * (1-falloff_color->a) + falloff_color->g * falloff_color->a) * glossy->g;
				result->b = (rcol.b * (1-falloff_color->a) + falloff_color->b * falloff_color->a) * glossy->b;
			}
			else													//no falloff
			{
				result->r = rcol.r * glossy->r;
				result->g = rcol.g * glossy->g;
				result->b = rcol.b * glossy->b;
			}
		}
		else if (falloff_stop > 0.f && falloff_to_color && falloff_color->a > 0.f)		//nothing hit, mix black color with falloff color by falloff.alpha
		{
			result->r = falloff_color->r * falloff_color->a * glossy->r;
			result->g =	falloff_color->g * falloff_color->a * glossy->g;
			result->b = falloff_color->b * falloff_color->a * glossy->b;
		}
	}
	else		//isotropic or anisotropic
	{
		if(samples > 1) // optimization for secondary traces, sampling decrease factor used
		{
			if(final_gather_mode(state))
				samples = 1;
			else
			{
				int level = state->refraction_level + state->reflection_level;
				samples >>= level*samples_decrease;
				if(samples < 1)
					samples = 1;
			}
		}
		miVector u, v;
		if(anisotropic)
			p_anis_orientation(&u, &v, state);

		double pSamples[2];
		int sample_number = 0;
		miUint samps = samples;

		while (mi_sample(pSamples, &sample_number, state, 2, &samps)) 
		{
			if (anisotropic)
				mi_reflection_dir_anisglossy_x(&rdir, state, &u, &v, *shiny_u, *shiny_v, pSamples);
			else if (*shiny)  // isotropic
				mi_reflection_dir_glossy_x(&rdir, state, *shiny, pSamples);

			if(mi_trace_reflection(&rcol, state, &rdir) || mi_trace_environment(&rcol, state, &rdir))
			{
				double dist = 0.f;
				if (state->child) // zero distance for environment calls
					dist = state->child->dist;
				if(falloff_stop > 0.f && dist && falloff_to_color) //falloff to color used
				{
					if(falloff_start < dist && falloff_stop > dist)
					{
	//					float atten = 1.f;
	//					atten = (float)exp(-(dist -falloff_start) / (falloff_stop - falloff_start));
						float atten = (float)(dist -falloff_start) / (falloff_stop - falloff_start);
						result->r += (rcol.r * (1-atten) + falloff_color->r * atten) * glossy->r;
						result->g += (rcol.g * (1-atten) + falloff_color->g * atten) * glossy->g;
						result->b += (rcol.b * (1-atten) + falloff_color->b * atten) * glossy->b;
					}
					else if (dist >= falloff_stop)		//full falloff color
					{
						result->r += falloff_color->r * glossy->r;
						result->g += falloff_color->g * glossy->g;
						result->b += falloff_color->b * glossy->b;
					}
					else if (dist <= falloff_start)		//no falloff color
					{
						result->r += rcol.r * glossy->r;
						result->g += rcol.g * glossy->g;
						result->b += rcol.b * glossy->b;
					}
				}
				else if (falloff_stop > 0.f && falloff_to_color)		//environment call with falloff color
				{
					result->r += (rcol.r * (1-falloff_color->a) + falloff_color->r * falloff_color->a) * glossy->r;
					result->g += (rcol.g * (1-falloff_color->a) + falloff_color->g * falloff_color->a) * glossy->g;
					result->b += (rcol.b * (1-falloff_color->a) + falloff_color->b * falloff_color->a) * glossy->b;
				}
				else													//no falloff
				{
					result->r += rcol.r * glossy->r;
					result->g += rcol.g * glossy->g;
					result->b += rcol.b * glossy->b;
				}
			}
			else if (falloff_stop > 0.f && falloff_to_color && falloff_color->a > 0.f)		//nothing hit, mix black color with falloff color by falloff.alpha
			{
				result->r += falloff_color->r * falloff_color->a * glossy->r;
				result->g += falloff_color->g * falloff_color->a * glossy->g;
				result->b += falloff_color->b * falloff_color->a * glossy->b;
			}
		}
		float invSamples = 1.f/samps;
		result->r *= invSamples;
		result->g *= invSamples;
		result->b *= invSamples; 
	}
	if(falloff_changed)
		mi_ray_falloff(state, &falloff_near, &falloff_far); //restore
	return miTRUE;
}

miBoolean p_refract(miColor *result, miState *state, p_MegaTK_t *paras, miColor *glossy, miScalar *shiny,
                    miBoolean anisotropic, miScalar *shiny_u, miScalar *shiny_v,
                    miColor *absorb_in, miColor *absorb_out, 
					miScalar falloff_start, miScalar falloff_stop, miBoolean falloff_to_color, miColor *falloff_color, 
					int samples, miCBoolean enter_surf, int samples_decrease)
{
	if(samples < 1)
		return miTRUE;

	float dist;
	miVector rdir;
	miColor  rcol;
	rdir.x = rdir.y = rdir.z = 0;
	rcol.r = rcol.g = rcol.b = rcol.a = 0;

	if(falloff_start > falloff_stop)
		falloff_start = falloff_stop;
	miBoolean falloff_changed = miFALSE;
	miGeoScalar falloff_near = falloff_start;
	miGeoScalar falloff_far = falloff_stop;

	if(falloff_stop > 0.f && !falloff_to_color)	//limit ray lenght
	{
		mi_ray_falloff(state, &falloff_near, &falloff_far);
		falloff_changed = miTRUE;
	}

//  miBoolean entering = (enter_surf != miFALSE);

   if(!*shiny && !anisotropic)
   {
		if(mi_refraction_dir(&rdir, state, state->ior_in, state->ior))
		if(mi_trace_refraction(&rcol, state, &rdir) || mi_trace_environment(&rcol, state, &rdir))
		{
            dist = 0.f;
            if (state->child) // zero distance for environment calls
               dist = (float)state->child->dist;
			if(enter_surf && dist)
				absorb(&rcol, absorb_in, dist);
			else if(dist)
				absorb(&rcol, absorb_out, dist);

			if(falloff_stop > 0.f && dist && falloff_to_color) //falloff to color used
			{
				if(falloff_start < dist && falloff_stop > dist)
				{
//					float atten = 1.f;
//					atten = (float)exp(-(dist -falloff_start) / (falloff_stop - falloff_start));
					float atten = (float)(dist -falloff_start) / (falloff_stop - falloff_start);
					result->r = (rcol.r * (1-atten) + falloff_color->r * atten) * glossy->r;
					result->g = (rcol.g * (1-atten) + falloff_color->g * atten) * glossy->g;
					result->b = (rcol.b * (1-atten) + falloff_color->b * atten) * glossy->b;
				}
				else if (dist >= falloff_stop)		//full falloff color
				{
					result->r = falloff_color->r * glossy->r;
					result->g = falloff_color->g * glossy->g;
					result->b = falloff_color->b * glossy->b;
				}
				else if (dist <= falloff_start)		//no falloff color
				{
					result->r = rcol.r * glossy->r;
					result->g = rcol.g * glossy->g;
					result->b = rcol.b * glossy->b;
				}
			}
			else if (falloff_stop > 0.f && falloff_to_color)		//environment call with falloff color
			{
				result->r = (rcol.r * (1-falloff_color->a) + falloff_color->r * falloff_color->a) * glossy->r;
				result->g =	(rcol.g * (1-falloff_color->a) + falloff_color->g * falloff_color->a) * glossy->g;
				result->b = (rcol.b * (1-falloff_color->a) + falloff_color->b * falloff_color->a) * glossy->b;
			}
			else													//no falloff
			{
				result->r = rcol.r * glossy->r;
				result->g = rcol.g * glossy->g;
				result->b = rcol.b * glossy->b;
			}
		}
		else if (falloff_stop > 0.f && falloff_to_color && falloff_color->a > 0.f)		//nothing hit, mix black color with falloff color by falloff.alpha
		{
			result->r = falloff_color->r * falloff_color->a * glossy->r;
			result->g =	falloff_color->g * falloff_color->a * glossy->g;
			result->b = falloff_color->b * falloff_color->a * glossy->b;
		}
   }
   else			//isotropic or anisotropic
   {
		if(samples > 1) // optimization for secondary traces
		{
			if(final_gather_mode(state))
				samples = 1;
			else
			{
				int level = state->refraction_level + state->reflection_level;
				samples >>= level*samples_decrease;
				if(samples < 1)
					samples = 1;
			}
		}
		miVector u, v;
		if (anisotropic)
			p_anis_orientation(&u, &v, state);

		double pSamples[2];
		int sample_number = 0;
		miUint samps = samples;

		while (mi_sample(pSamples, &sample_number, state, 2, &samps)) 
		{
			if (anisotropic)
			{
				if(!mi_transmission_dir_anisglossy_x(&rdir, state, state->ior_in, state->ior, &u, &v, *shiny_u, *shiny_v, pSamples))
				continue; // if we are not in snell's window, refraction is black anyway, so no need to trace
			} 
			else if (*shiny)  // isotropic
			{
				if(!mi_transmission_dir_glossy_x(&rdir, state, state->ior_in, state->ior, *shiny, pSamples))
					continue; // if we are not in snell's window, refraction is black anyway, so no need to trace
			} 
			if(mi_trace_refraction(&rcol, state, &rdir) || mi_trace_environment(&rcol, state, &rdir))
			{
				dist = 0.f;
				if (state->child) // zero distance for environment calls
				dist = (float)state->child->dist;

				if(enter_surf && dist)
					absorb(&rcol, absorb_in, dist);
				else if(dist)
					absorb(&rcol, absorb_out, dist);

				if(falloff_stop > 0.f && dist && falloff_to_color) //falloff to color used
				{
					if(falloff_start < dist && falloff_stop > dist)
					{
	//					float atten = 1.f;
	//					atten = (float)exp(-(dist -falloff_start) / (falloff_stop - falloff_start));
						float atten = (float)(dist -falloff_start) / (falloff_stop - falloff_start);
						result->r += (rcol.r * (1-atten) + falloff_color->r * atten) * glossy->r;
						result->g += (rcol.g * (1-atten) + falloff_color->g * atten) * glossy->g;
						result->b += (rcol.b * (1-atten) + falloff_color->b * atten) * glossy->b;
					}
					else if (dist >= falloff_stop)		//full falloff color
					{
						result->r += falloff_color->r * glossy->r;
						result->g += falloff_color->g * glossy->g;
						result->b += falloff_color->b * glossy->b;
					}
					else if (dist <= falloff_start)		//no falloff color
					{
						result->r += rcol.r * glossy->r;
						result->g += rcol.g * glossy->g;
						result->b += rcol.b * glossy->b;
					}
				}
				else if (falloff_stop > 0.f && falloff_to_color)		//environment call with falloff color
				{
					result->r += (rcol.r * (1-falloff_color->a) + falloff_color->r * falloff_color->a) * glossy->r;
					result->g += (rcol.g * (1-falloff_color->a) + falloff_color->g * falloff_color->a) * glossy->g;
					result->b += (rcol.b * (1-falloff_color->a) + falloff_color->b * falloff_color->a) * glossy->b;
				}
				else													//no falloff
				{
					result->r += rcol.r * glossy->r;
					result->g += rcol.g * glossy->g;
					result->b += rcol.b * glossy->b;
				}
			}
			else if (falloff_stop > 0.f && falloff_to_color && falloff_color->a > 0.f)		//nothing hit, mix black color with falloff color by falloff.alpha
			{
				result->r += falloff_color->r * falloff_color->a * glossy->r;
				result->g += falloff_color->g * falloff_color->a * glossy->g;
				result->b += falloff_color->b * falloff_color->a * glossy->b;
			}
		}
		float invSamples = 1.f/samps;
		result->r *= invSamples;
		result->g *= invSamples;
		result->b *= invSamples;
	}
	if(falloff_changed)
		mi_ray_falloff(state, &falloff_near, &falloff_far); //restore
	return miTRUE;
}

miBoolean p_path_diffuse(miColor *result, miState *state, int max_depth, int samples, int samples_decrease)
{
	result->r = result->g = result->b = result->a = 0;
//	if((state->reflection_level >= state->options->reflection_depth) || (state->reflection_level >= max_depth))
	if(state->reflection_level >= state->options->reflection_depth)
		return(miFALSE);
	if((samples < 1) || (max_depth < 1))
		return(miTRUE);

	int *rayDuffuse;
	int curDiffuse;
//	int size_rayDuffuse = 0;
	bool first_state = false;

	miVector rdir;
	miColor  rcol;
	rdir.x = rdir.y = rdir.z = 0;
	rcol.r = rcol.g = rcol.b = rcol.a = 0;

	if(state->reflection_level >= 1)
	{
		rayDuffuse = (int *)mi_shaderstate_get(state, pathDepth_key, NULL);  //"rad" used for count reflect_diffuse depth
		if (rayDuffuse != NULL)
		{
			curDiffuse = *rayDuffuse;
			curDiffuse++;
		}
		else //initialise it.	
		{
			first_state = true;		//first state, set "rad"
			curDiffuse = 1;
		}
	}
	else //it couldn't be any existing state because current reflection depth = 0, set first_state
	{
		first_state = true;		//first state, set "rad"
		curDiffuse = 1;
	}

	if(curDiffuse > max_depth)  //check for max trace depth
		return(miTRUE);

	if(samples > 1) // optimization for secondary traces
	{
		if(first_state == false) //it's duffuse reflection call, set samples to 1 
			samples = 1;
		else											//it's first state or not diffuse call, increase samples by current depth
		{
			if(final_gather_mode(state))
				samples >>= 3;					//samples = samples/8
			else
				samples >>= (state->refraction_level + state->reflection_level)*samples_decrease;
			if(samples < 1)
				samples = 1;
		}		
	}

	if(samples > 1)
	{
		double pSamples[2];
		int sample_number = 0;
		miUint samps = samples;
		while (mi_sample(pSamples, &sample_number, state, 2, &samps)) 
		{
			mi_shaderstate_set(state, pathDepth_key, &curDiffuse, sizeof(curDiffuse), miSS_LIFETIME_EYERAY);
			mi_reflection_dir_diffuse(&rdir, state);
			if((mi_trace_reflection(&rcol, state, &rdir) && state->child->pri) || mi_trace_environment(&rcol, state, &rdir))
			{
				result->r += rcol.r;
				result->g += rcol.g;
				result->b += rcol.b;
			}
		}
//		float invSamples = 1.f/samps;
		result->r /= samps;
		result->g /= samps;
		result->b /= samps; 
	}
	else		//only one sample
	{
		mi_shaderstate_set(state, pathDepth_key, &curDiffuse, sizeof(curDiffuse), miSS_LIFETIME_EYERAY);
		mi_reflection_dir_diffuse(&rdir, state);
		if((mi_trace_reflection(&rcol, state, &rdir) && state->child->pri) || mi_trace_environment(&rcol, state, &rdir))
		{
			result->r = rcol.r;
			result->g = rcol.g;
			result->b = rcol.b;
		}
	}
	if(first_state)				//null state only for first state
		mi_shaderstate_set(state, pathDepth_key, NULL, sizeof(NULL), miSS_LIFETIME_EYERAY);

	return miTRUE;
}
