//  Copyright (c) 2005, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"
#include "geoshader.h"
#include "p_utils.h"

typedef struct
{
	miBoolean	enable_wrap_diffuse;
	miScalar	diffuse;
	miColor		wrap_color;		
	miScalar	wrap;
	miScalar	term_bright;
	miColor		term_color;
	miScalar	term_width;
	miBoolean	enable_scatter;
	miScalar	scatter;
	miColor		scatter_color;
	miScalar	max_dist_scatter;
	miScalar	spread;
	int			samples;
	miBoolean	shadow;
	miScalar	shadow_blur;
	miScalar	shadow_transp;
	miBoolean	enable_translucence;
	miScalar	translucence;
	miColor		translucence_color;
	int			mode;		/* light mode: 0..2 */
	int			i_light;	/* index of first light */
	int			n_light;	/* number of lights */
	miTag		lights[1];	/* list of lights */
}p_scatter_wrap_t;

extern "C"
{
DLLEXPORT int p_scatter_wrap_version(void) {return(3);}

DLLEXPORT miBoolean p_scatter_wrap(miColor *result, miState *state, p_scatter_wrap_t *paras)
{
	if(state->type == miRAY_SHADOW || state->type == miRAY_DISPLACE)
		return miFALSE;

	miTag		*light;			// tag of light instance
	int		n_l;				// number of light sources
	int		i_l;				// offset of light sources
	int		m;					// light mode: 0=all, 1=incl, 2=excl
	int			samples;		// # of samples taken
	miBoolean	shadow;			//shadow (depth 2)
	miScalar	sh_blur;
	miColor		color;			// color from light source
	miScalar	sum;			// summed sample colors
	miColor		sum_col;		// summed sample colors
	miVector	dir;			// direction towards light
	miScalar	dot_nl;			// dot prod of normal and dir
	miScalar	geom_dot_nl = 0.f;			// dot prod of normal and dir]
	miScalar blend_normal = 1.0f;	// blend light direction with normal

	double		sample[2];		// 2d sample, will be filled by mi_sample
	miInteger	sampleCnt;		// sample counter, needed for mi_sample
	miVector	sampleDir;		// unit direction vector ray will be shot at, and the current normal n
	
	miBoolean enable_wrap_diffuse = *mi_eval_boolean(&paras->enable_wrap_diffuse);
	miBoolean enable_scatter = *mi_eval_boolean(&paras->enable_scatter);
	miScalar diffuse		= *mi_eval_scalar(&paras->diffuse);
	miScalar max_scat_dist	= *mi_eval_scalar(&paras->max_dist_scatter);
	miScalar wrap			= *mi_eval_scalar(&paras->wrap);
	miScalar scatter_coef	= *mi_eval_scalar(&paras->scatter);
	miUint scatter_samples	= *mi_eval_integer(&paras->samples);
	miScalar spread			= *mi_eval_scalar(&paras->spread);
	miScalar term_width		= (1-*mi_eval_scalar(&paras->term_width));
	miScalar term_bright	= *mi_eval_scalar(&paras->term_bright);
	miScalar light_dot_nl = 0;			//modified dot_nl for teminator
	shadow = *mi_eval_boolean(&paras->shadow);
	sh_blur = *mi_eval_scalar(&paras->shadow_blur);
	miScalar shadow_transp = *mi_eval_scalar(&paras->shadow_transp);

	miScalar translucence			= *mi_eval_scalar(&paras->translucence);
	miBoolean enable_translucence	= *mi_eval_boolean(&paras->enable_translucence);
	miColor *translucence_color;
	if(enable_translucence)
		translucence_color = mi_eval_color(&paras->translucence_color);

	miColor wrap_color = {0,0,0,0};
	miColor term_color = {0,0,0,0};
	if(enable_wrap_diffuse && diffuse)
	{
		if(diffuse)
			wrap_color = *mi_eval_color(&paras->wrap_color);
		if(term_bright)
			term_color = *mi_eval_color(&paras->term_color);
	}
	miColor scatter_color = {0,0,0,0};
	if(enable_scatter && scatter_coef)
		scatter_color = *mi_eval_color(&paras->scatter_color);

	m  = *mi_eval_integer(&paras->mode);

	result->r = result->g = result->b = result->a = 0;
	sampleCnt = 0;
	miScalar scatter = 0;

	miColor transl_color;
//	miColor light_color;

	miVector orig_normal = state->normal;

	miOptions *orig_options = (miOptions *)(state->options);	//save pointer to original options
	miOptions m_options = *state->options;		//copy original options
	state->options = &m_options;				//link new options to state
	m_options.shadow = 0;					//edit new options
//	state->normal.x = state->normal.y = state->normal.z = 0;	//edit normal

//	miRc_intersection *pri = state->pri;
//	state->pri = NULL;

	miScalar cur_dist = 0;
	miScalar light_distance = 0;
	miVector inv_normal;
	miVector ldir;
	miTag ltag;
	miTag cur_light;
	int light_type;

	n_l   = *mi_eval_integer(&paras->n_light);
	i_l   = *mi_eval_integer(&paras->i_light);
	light =  mi_eval_tag(paras->lights) + i_l;

	if (m == 1)		// modify light list (inclusive mode)
		mi_inclusive_lightlist(&n_l, &light, state);
	else if (m == 2)	// modify light list (exclusive mode)
		mi_exclusive_lightlist(&n_l, &light, state);
	else if (m == 4) 
	{
		n_l = 0;
		light = 0;
	}

	// Loop over all light sources
	if (m==4 || n_l) 
	{
		state->normal.x = state->normal.y = state->normal.z = 0;
		if(enable_scatter || enable_wrap_diffuse)
		{
			for (mi::shader::LightIterator iter(state, light, n_l); !iter.at_end(); ++iter) 
			{
				sum_col.r = sum_col.g = sum_col.b = sum_col.a = 0;			
				cur_light = iter->get_light_tag(iter->get_current());
				while (iter->sample())
				{
					state->normal = orig_normal;
					dot_nl = iter->get_dot_nl();
					iter->get_contribution(&color);
					dir = iter->get_direction();				
					sum = 0;				
					light_distance = (miScalar)(state->child->dist);
					mi_vector_normalize(&dir);
					mi_vector_normalize(&orig_normal);
					dot_nl = mi_vector_dot(&dir, &orig_normal);
					//Compute scatter
					if(enable_scatter && scatter_coef && scatter_samples && max_scat_dist)
					{
						if(spread == 0)
						{
							scatter = 0;
							inv_normal = orig_normal;
							mi_vector_neg(&inv_normal);	
		//					mi_vector_mul(&inv_normal, blend_normal);
							mi_vector_add(&sampleDir, &inv_normal, &dir);	// add normal direction
							mi_vector_normalize(&sampleDir);
							
							if(mi_trace_probe(state, &sampleDir, &state->point))
							{
								if(state->child->inv_normal)
								{
									mi_query(miQ_INST_ITEM, state, cur_light, &ltag);
									light_type = -1;
									if(!mi_query(miQ_LIGHT_TYPE, state, ltag, &light_type))
										mi_warning("mi_query(miQ_LIGHT_TYPE) failed!");

									cur_dist = (miScalar)(state->child->dist);
									if(cur_dist > light_distance && light_type != 1)			//light found inside object
										cur_dist = light_distance;	
									if(cur_dist > max_scat_dist)
										cur_dist = max_scat_dist;
									cur_dist /= max_scat_dist;

									if(shadow && shadow_transp < 1)						//Shadow (Depth = 2)
									{
										//Trace additional prob ray toward light for shadow occlusion
										miState *childState;
										miState *grandChildState = new miState;
										childState = state->child;
										childState->child = grandChildState;

										if(light_type == 1)		//direct light without origin, just take geom direction
											mi_query(miQ_LIGHT_DIRECTION, childState, cur_light, &ldir);
										else if(light_type == 0 || light_type == 2)	//other lights type with origin
										{
											mi_query(miQ_LIGHT_ORIGIN, childState, cur_light, &ldir);
											mi_vector_sub(&ldir, &childState->point, &ldir);
										} else
											mi_warning("p_scatter_wrap: Unsupported light type! Light type is = %i", light_type);
										mi_vector_neg(&ldir);
										if(sh_blur > 0)
										{
											mi_vector_normalize(&ldir);
											//Shift vector to light
											double rand_value[3];
											mi_sample(rand_value, 0, childState, 3, 0);

											miVector shiftVL;				//random vector from -sh_blur to sh_blur value	
											shiftVL.x = (miScalar)(sh_blur * (2 * rand_value[0] -1));
											shiftVL.y = (miScalar)(sh_blur * (2 * rand_value[1] -1));
											shiftVL.z = (miScalar)(sh_blur * (2 * rand_value[2] -1));

											mi_vector_add(&ldir, &ldir, &shiftVL);		//mix random vector and light direction
										}
										mi_vector_normalize(&ldir);
										if(mi_trace_probe(childState, &ldir, &childState->point))	//trace probe for shadow
										{
											if((miScalar)(state->child->dist) + (miScalar)(childState->child->dist) > light_distance && light_type != 1)		//intersection found behind the light
											{
												if(dot_nl > 0)												
													scatter += (1-cur_dist) * (1-dot_nl);
												else
													scatter += (1-cur_dist);
											}
											else if(shadow_transp > 0)
											{
												if(dot_nl > 0)												
													scatter += shadow_transp * (1-cur_dist) * (1-dot_nl);
												else
													scatter += shadow_transp * (1-cur_dist);												
											}
										}
										else	
										{
											if(dot_nl > 0)												
												scatter += (1-cur_dist) * (1-dot_nl);
											else
												scatter += (1-cur_dist);
										}
										delete grandChildState;
									}
									else //shadow
									{
										if(dot_nl > 0)												
											scatter += (1-cur_dist) * (1-dot_nl);
										else
											scatter += (1-cur_dist);
									}						
									sum += scatter * scatter_coef;
									
								}
							}
						}
						else		//spread > 0
						{
							scatter = 0;
							sampleCnt = 0;
							mi_vector_mul(&dir, 1.0f - (spread / 180.0f));					
							while(mi_sample(sample, &sampleCnt, state, 2, &scatter_samples))
							{
								sampleDir = p_sphere_sample(sample);		// map the sample to a sphere
								mi_vector_mul(&sampleDir, spread / 180.0f);		// premultiply this
								mi_vector_add(&sampleDir, &dir, &sampleDir);	// add it together
								inv_normal = orig_normal;
								mi_vector_neg(&inv_normal);
		//						mi_vector_mul(&inv_normal, blend_normal);
								mi_vector_add(&sampleDir, &inv_normal, &sampleDir);	// add normal direction
								mi_vector_normalize(&sampleDir);

								if(mi_trace_probe(state, &sampleDir, &state->point))
								{
									if(state->child->inv_normal)
									{
										mi_query(miQ_INST_ITEM, state, cur_light, &ltag);
										light_type = -1;
										if(!mi_query(miQ_LIGHT_TYPE, NULL, ltag, &light_type))
											mi_warning("mi_query(miQ_LIGHT_TYPE) failed!");

										cur_dist = (miScalar)(state->child->dist);
										if(cur_dist > light_distance && light_type != 1)			//light found inside object
											cur_dist = light_distance;	
										if(cur_dist > max_scat_dist)
											cur_dist = max_scat_dist;
										cur_dist /= max_scat_dist;

										if(shadow && shadow_transp < 1)						//Depth = 2 (shadow effect)
										{
											//Trace additional prob ray toward light for shadow occlusion
											miState *childState;
											miState *grandChildState = new miState;
											childState = state->child;
											childState->child = grandChildState;

											if(light_type == 1)		//direct light without origin, just take geom direction
												mi_query(miQ_LIGHT_DIRECTION, childState, cur_light, &ldir);
											else if(light_type == 0 || light_type == 2)					//other lights type with origin
											{											
												mi_query(miQ_LIGHT_ORIGIN, childState, cur_light, &ldir);
												mi_vector_sub(&ldir, &childState->point, &ldir);
											} else
												mi_warning("p_scatter_wrap: Unsupported light type! Light type is = %i", light_type);
											mi_vector_neg(&ldir);
											if(sh_blur > 0)
											{
												mi_vector_normalize(&ldir);
												//Shift vector to light
												double rand_value[3];
												mi_sample(rand_value, 0, childState, 3, 0);

												miVector shiftVL;				//random vector from -sh_blur to sh_blur value	
												shiftVL.x = (miScalar)(sh_blur * (2 * rand_value[0] -1));
												shiftVL.y = (miScalar)(sh_blur * (2 * rand_value[1] -1));
												shiftVL.z = (miScalar)(sh_blur * (2 * rand_value[2] -1));

												mi_vector_add(&ldir, &ldir, &shiftVL);		//mix random vector and light direction
											}
											mi_vector_normalize(&ldir);
											if(mi_trace_probe(childState, &ldir, &childState->point))	//trace probe for shadow
											{
												if((miScalar)(state->child->dist) + (miScalar)(childState->child->dist) > light_distance && light_type != 1)		//intersection found behind the light
												{
													if(dot_nl > 0)												
														scatter += (1-cur_dist) * (1-dot_nl);
													else
														scatter += (1-cur_dist);
												}
												else if(shadow_transp > 0)
												{													
													if(dot_nl > 0)												
														scatter += shadow_transp * (1-cur_dist) * (1-dot_nl);
													else
														scatter += shadow_transp * (1-cur_dist);													
												}
											}
											else
											{
												if(dot_nl > 0)												
													scatter += (1-cur_dist) * (1-dot_nl);
												else
													scatter += (1-cur_dist);
											}
											delete grandChildState;
										}
										else
										{
											if(dot_nl > 0)												
												scatter += (1-cur_dist) * (1-dot_nl);
											else
												scatter += (1-cur_dist);
										}
									}
								}
							}
							sum += scatter / sampleCnt * scatter_coef;
						
						}
						sum_col.r += scatter_color.r * sum;
						sum_col.g += scatter_color.g * sum;
						sum_col.b += scatter_color.b * sum;
						sum = 0;
					}
					//Wrap
					if(enable_wrap_diffuse)
					{
						//Light terminator
						if(term_bright)
						{
							if(dot_nl >= 0)
								light_dot_nl = 1 -dot_nl;
							if(dot_nl < 0)
								light_dot_nl = 1 + dot_nl;

							light_dot_nl = (light_dot_nl - term_width) / (1.0f - term_width);
							if(light_dot_nl < 0)
								light_dot_nl = 0;
							sum = light_dot_nl * term_bright;

							sum_col.r += term_color.r * sum;
							sum_col.g += term_color.g * sum;
							sum_col.b += term_color.b * sum;
							sum = 0;
						}
						//Compute diffuse
						if(diffuse)
						{
							dot_nl = (dot_nl + wrap) / (1.0f + wrap);		//wrap
							if(dot_nl > 0)
								sum = dot_nl * diffuse;

							sum_col.r += wrap_color.r * sum;
							sum_col.g += wrap_color.g * sum;
							sum_col.b += wrap_color.b * sum;
		//					sum = 0;
						}
					}  //wrap
					sum_col.r *= color.r;
					sum_col.g *= color.g;
					sum_col.b *= color.b;
				} //while
				samples = iter->get_number_of_samples();
				if(samples)
				{
					result->r += sum_col.r / samples;
					result->g += sum_col.g / samples;
					result->b += sum_col.b / samples;
				}
				state->normal.x = state->normal.y = state->normal.z = 0;
				
			} //for LightIterator
		} //if sattering and wrap
		
		//Translucency
		if(enable_translucence && translucence > miEPS && translucence_color->r > miEPS && translucence_color->g > miEPS && translucence_color->b > miEPS) //Translucency
		{
			m_options.shadow = 1;
			transl_color.r = transl_color.g = transl_color.b = 0.f;
			for(mi::shader::LightIterator iter(state, light, n_l); !iter.at_end(); ++iter)
			{
				inv_normal = orig_normal;
				mi_vector_neg(&inv_normal);
				state->normal = inv_normal;
				transl_color.r = transl_color.g = transl_color.b = 0.f;
				while (iter->sample())
				{
					dir = iter->get_direction();	
					dot_nl = mi_vector_dot(&dir, &orig_normal);
					geom_dot_nl = mi_vector_dot(&dir, &state->normal_geom);
					if(dot_nl < 0 && geom_dot_nl < 0)
					{
						dot_nl = -dot_nl;
						dot_nl *= dot_nl;
						iter->get_contribution(&color);
						dot_nl *= translucence;
						transl_color.r += color.r * dot_nl * translucence_color->r;
						transl_color.g += color.g * dot_nl * translucence_color->g;
						transl_color.b += color.b * dot_nl * translucence_color->b;
					}
				}
				samples = iter->get_number_of_samples();
				if(samples)
				{
					result->r += transl_color.r / samples;
					result->g += transl_color.g / samples;
					result->b += transl_color.b / samples;
				}
				state->normal = orig_normal;
			}
			state->normal.x = state->normal.y = state->normal.z = 0;
			m_options.shadow = 0;
		} // if translucence
		
	} //if
	state->normal = orig_normal;
	state->options = orig_options;				//restore pointer to original options
	return(miTRUE);
}
}
