//  Copyright (c) 2007, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"
#include "geoshader.h"

#define EPS	 1e-4
#define BLACK(C) ((C).r==0 && (C).g==0 && (C).b==0)

typedef struct {
		miColor		color;			// RGBA material color
		miScalar	transp;			// float transparency
		miBoolean	inverse_transp;	//
		miBoolean	atten;
		miScalar	start;
		miScalar	stop;
		miScalar	stop_transp;	// shadow trasparency after "stop" distance
		miScalar	edginess;		// fresnell effect
		int		mode;				// 0=all, 1=inclusive, 2=exclusive
		int		i_light;			// index of first light
		int		n_light;			// number of lights
		miTag		light[1];		// list of lights
}p_shadow_transparency_t;

static void choose_volume(miState *state)
{
	register int		n = 0;				// same-material counter
	register miState	*s, *s_in = 0;		// for finding enclosing mtl

	for (n=0, s=state; s; s=s->parent)		// history?
		if ((s->type == miRAY_TRANSPARENT	   ||
		     s->type == miRAY_REFRACT		   ||
		     s->type == miPHOTON_TRANSMIT_SPECULAR ||
		     s->type == miPHOTON_TRANSMIT_GLOSSY   ||
		     s->type == miPHOTON_TRANSMIT_DIFFUSE) &&
		     s->parent && s->parent->shader == state->shader)
		{
			n++;
			if (!s_in)
				s_in = s->parent;
		}

	if (n & 1)					// odd: exiting
		state->refraction_volume = s_in ? s_in->volume : state->camera->volume;
	else						// even: entering
		if (!state->refraction_volume)
			state->refraction_volume = state->volume;
}

extern "C"
{
DLLEXPORT int p_shadow_transparency_version(void) {return(1);}

DLLEXPORT miBoolean p_shadow_transparency(miColor *result, miState *state, p_shadow_transparency_t *paras)
{
		int			n, i, mode;
		miTag			*light;
		miScalar		transp;
		miColor	color;
		miScalar start;
		miScalar stop;
		miScalar stop_transp;
		miScalar edginess;

		if (state->type != miRAY_SHADOW)
			return(miFALSE);

	//	result->r = result->g = result->b = 1;

		float distance = (miScalar)state->dist;

		mode = *mi_eval_integer(&paras->mode);
		if(mode < 3)
		{
			n = *mi_eval_integer(&paras->n_light);
			light = mi_eval_tag(paras->light) +	*mi_eval_integer(&paras->i_light);
			switch(mode)
			{
				case 1: mi_inclusive_lightlist(&n, &light, state); break;
				case 2: mi_exclusive_lightlist(&n, &light, state); break;
			}
			for (i=0; i < n; i++, light++)
				if (state->light_instance == *light)
					break;
			if (i == n)
				return(miTRUE);
		}

		color  =  *mi_eval_color(&paras->color);
		transp = *mi_eval_scalar(&paras->transp);
		start = *mi_eval_scalar(&paras->start);
		stop = *mi_eval_scalar(&paras->stop);
		stop_transp = *mi_eval_scalar(&paras->stop_transp);
		edginess = *mi_eval_scalar(&paras->edginess);

		if(*mi_eval_scalar(&paras->inverse_transp))		//inverse transparency
			transp = 1-transp;

		if(transp > 1)									//clamp transparency
			transp = 1;
		else if (transp < 0)
			transp = 0;

		if(stop_transp < transp)
			stop_transp = transp;

		if (state->options->shadow == 's')	//segment shadow
		{
			if(edginess && transp)					//use edgness
			{
				miVector V = state->dir;
				miVector N = state->normal;
				mi_vector_neg(&V);
				mi_vector_normalize(&V);
				mi_vector_normalize(&N);
				miScalar edge = mi_vector_dot(&V, &N);
				if(edge < 0)
					edge *= -1;
				transp = transp * edge * edginess + transp * (1-edginess);
			}
			if(*mi_eval_boolean(&paras->atten))		//compute attenuation gradient
			{
				if(distance <= start)
					distance = 0;
				else
				{
					if(distance >= stop)
						distance = 1;
					else
						distance = (distance - start)/(stop - start);
				}
			}
			if(transp > 0.0001)
			{
				choose_volume(state);
				mi_trace_shadow_seg(result, state);
			}
			//posible bug HERE!!!!!  maybe need remove '+ transp'
			result->r *= color.r * (1-transp) + transp;
			result->g *= color.g * (1-transp) + transp;
			result->b *= color.b * (1-transp) + transp;

			if(distance && *mi_eval_boolean(&paras->atten) && (state->parent->type != miRAY_SHADOW)) //apply attenuation
			{
				distance *= stop_transp;
				result->r = result->r * (1-distance) + distance;
				result->g = result->g * (1-distance) + distance;
				result->b = result->b * (1-distance) + distance;
			}
		}
		else			//sort or simple
		{
			if(edginess && transp)					//use edgness
			{
				miVector V = state->dir;
				miVector N = state->normal;
				mi_vector_neg(&V);
				mi_vector_normalize(&V);
				mi_vector_normalize(&N);
				miScalar edge = mi_vector_dot(&V, &N);
				if(edge < 0)
					edge = -edge;
				transp = transp * edge * edginess + transp * (1-edginess);
			}
			result->r *= color.r * (1-transp) + transp;
			result->g *= color.g * (1-transp) + transp;
			result->b *= color.b * (1-transp) + transp;
		}
//		mi_info("%f", distance);
		return(miBoolean)(result->r || result->g || result->b);
}
}
