#pragma once

#ifndef p_utils_h
#define p_utils_h

#include <limits.h>

// 1/sqrt(2*pi)
#define M_SQRT_1_2_PI 0.39894228040143267793994605993438

#define miAlphaEPS 0.001
#define miAlphaEPS_INV 0.999
#define miEPS 0.0001
#define miEPS_INV 0.9999

#define RTA 180/M_PI	//radian to angle
#define ATR M_PI/180	//angle to radian
#define ITM 25.400051f  //inch to mm

#define M_2PI 2.0*M_PI

//#define USHRT_MAX 65535

#define lightsTK_key "ltk"		//struct key returned by TK lights
#define pathDepth_key "rad"		//struct key for path tracing depth

inline miBoolean final_gather_mode(miState *state)
{
    int fg, tile[4];
    //If miQ_TILE_PIXELS returns false, we are in the precompute stage
    if (!mi_query(miQ_TILE_PIXELS, state, miNULLTAG, tile))
        return miTRUE;
    if (state->type == miRAY_FINALGATHER) 
        return miTRUE;
    mi_query(miQ_FINALGATHER_STATE, state, miNULLTAG, &fg);
    if (fg) return miTRUE;
    return miFALSE;
}

// clamp a scalar from 0 to 1
#define p_clamp_0_1(_value)						\
	((_value) > 1.0f) ? 1.0f : ( ((_value) < 0.0f) ? 0.0f : _value )

// return maximum of a and b
#define p_max(_a, _b)								\
	((_a > _b)?_a:_b)

// return square of a
#define p_squared(_a)								\
	((_a) * (_a))

inline float
p_vector_length(miVector *vec)
{
	return sqrt(vec->x * vec->x + vec->y * vec->y + vec->z * vec->z);
}

// add two colors
static inline void 
p_color_add(miColor *r, miColor *a, miColor *b)
{
	r->r = a->r + b->r;
	r->g = a->g + b->g;
	r->b = a->b + b->b;
	r->a = a->a + b->a;
}

// multiply two colors
static inline void 
p_color_mul(miColor *r, miColor *a, miColor *b)
{
	r->r = a->r * b->r;
	r->g = a->g * b->g;
	r->b = a->b * b->b;
	r->a = a->a * b->a;
}

// mix two colors a and b based on a scalar c, where 0 means full color a, and
// 1 means full color b gets applied
static inline void 
p_color_mix(miColor *r, miColor *a, miColor *b, miScalar c)
{
	r->r = a->r * (1.0f - c) + b->r * c;
	r->g = a->g * (1.0f - c) + b->g * c;
	r->b = a->b * (1.0f - c) + b->b * c;
	r->a = a->a * (1.0f - c) + b->a * c;
}

// mix two colors a and b based on a third color c, where 0 means full color a, and
// 1 means full color b gets applied
static inline void 
p_color_mix(miColor *r, miColor *a, miColor *b, miColor *c)
{
	r->r = a->r * (1.0f - c->r) + b->r * c->r;
	r->g = a->g * (1.0f - c->r) + b->g * c->g;
	r->b = a->b * (1.0f - c->g) + b->b * c->b;
	r->a = a->a * (1.0f - c->a) + b->a * c->a;
}


// get the minimum of all color components
static inline miScalar
p_color_mincomp(miColor *r, miBoolean usealpha = miFALSE)
{
	register miScalar	minComp = r->r;

	if (minComp > r->g)
		minComp = r->g;

	if (minComp > r->b)
		minComp = r->b;

	if (usealpha && minComp > r->a)
		minComp = r->a;

	return minComp;
}


// get the maximum of all color components
static inline miScalar
p_color_maxcomp(miColor *r, miBoolean usealpha = miFALSE)
{
	register miScalar maxComp = r->r;

	if (maxComp < r->g)
		maxComp = r->g;

	if (maxComp < r->b)
		maxComp = r->b;

	if (usealpha && maxComp < r->a)
		maxComp = r->a;

	return maxComp;
}


// return a smoothstep interpolant of t between the borders a and b
static inline float p_smooth_step(float min, float max, float x)
{
	if(min != max)
	{
		float p, term;
		if (x <= min)
			p = min;
		else if (x >= max)
			p = max;
		else
			p = x;
		/* now calculate smoothstep value:
		   -2term^3 + 3term^2, where term = (p-min)/(max-min) */
		term  = (p - min) / (max - min);
		return -2 * term * term * term + 3 * term * term;
	}
	else
		return min;
}


// find the correct volume shader to stuff into the refraction volume
static inline void 
p_choose_volume(miState	*state)
{
	register int		n = 0;		/* same-material counter */
	register miState	*s, *s_in = 0;	/* for finding enclosing mtl */

	for (n=0, s=state; s; s=s->parent)		/* history? */
		if ((s->type == miRAY_TRANSPARENT	   ||
		     s->type == miRAY_REFRACT		   ||
		     s->type == miPHOTON_TRANSMIT_SPECULAR ||
		     s->type == miPHOTON_TRANSMIT_GLOSSY   ||
		     s->type == miPHOTON_TRANSMIT_DIFFUSE) &&
		     s->parent && s->parent->shader == state->shader) {
			n++;
			if (!s_in)
				s_in = s->parent;
		}

	if (n & 1)					/* odd: exiting */
		state->refraction_volume = s_in ? s_in->volume
						: state->camera->volume;
	else						/* even: entering */
		if (!state->refraction_volume)
			state->refraction_volume = state->volume;
}

static inline miVector
p_sphere_sample(miGeoScalar sample[2])
{
	// Compute a unit direction vector from 2 samples on a spherical hull:
	miVector	smp;
	double		t, st, phirad;

	t = 2.0 * sample[0] - 1.0;			// map from [0, 1) to [-1, 1)
	st = sqrt(1.0 - t*t);
	phirad = sample[1] * M_2PI;	// map from [0, 1) to [0, 2pi)
	smp.x = (miScalar)(st * cos(phirad));
	smp.y = (miScalar)(st * sin(phirad));
	smp.z = (miScalar)t;

	return smp;
}


// return fresnel reflectance depending on the IORs n1 and n2, and the
// cosinus of the incident angle, costheta
static inline float
p_fresnel(miScalar n1, miScalar n2, miScalar costheta)
{
	if(n2 != 0)
	{
		float n = n1/n2;
		float c = costheta * n;
		float g = (float)sqrt(1.0f + c*c - n*n);

		return 0.5f * ( p_squared((g-c) / (g+c)) * ( 1.0f + p_squared( (c*(g+c) - n*n) / (c*(g-c) + n*n) ) ) );
	}
	else
		return 0.0f;
}


// compute unit integral gaussian distribution at location x, 
// with standard deviation sigma
static inline miScalar
p_gauss_dist(miScalar sigma, miScalar x)
{
	if(sigma != 0)
		return (miScalar)((M_SQRT_1_2_PI * (1.0f / sigma)) * exp((x*x)/ (-2.0 * sigma * sigma)));
	else
		return 0.0f;
}


// same as above, but normalized
static inline miScalar
p_gauss_dist_norm(miScalar sigma, miScalar x)
{
	if(sigma != 0)
		return (miScalar)(p_gauss_dist(sigma, x) / (M_SQRT_1_2_PI * (1.0f / sigma)));
	else
		return 0.0f;
}


static void mib_lightmap_combine_vectors(
	miVector	*res,
	miVector const	*a,
	miVector const	*b,
	miVector const	*c,
	miVector const	*bary)
{
	res->x = bary->x * a->x + bary->y * b->x + bary->z * c->x;
	res->y = bary->x * a->y + bary->y * b->y + bary->z * c->y;
	res->z = bary->x * a->z + bary->y * b->z + bary->z * c->z;
} 

inline miRay_type p_choose_simple_scatter_type(miState *state, miScalar *transparent, miColor *diffuse, miColor *reflect, miColor *refract)
{
	float diff = 0;
	float refl = 0;
	float refr = 0;
	float transp = 0;

	if(diffuse)	diff = (diffuse->r + diffuse->g + diffuse->b) * 0.33333f;
	if(reflect)	refl = (reflect->r + reflect->g + reflect->b) * 0.33333f;
	if(refract) refr = (refract->r + refract->g + refract->b) * 0.33333f;
	if(transparent)	transp = *transparent;
	
	float average = diff + transp + refl + refr;  //in ideal it should be 1.0f

	if(average > miEPS)			//normalize 0-1
	{
		diff /= average;
		transp /= average;
		refl /= average;
		refl /= average;
	}
	else
		return(miPHOTON_ABSORB);

//	miScalar rand_value = mi_random();
	double rand_value;
	mi_sample(&rand_value, 0, state, 1, 0);

	if(rand_value <= diff && diff)							//diffuse reflect
	{
		diffuse->r = diffuse->g = diffuse->b = average;
		return(miPHOTON_REFLECT_DIFFUSE);
	}
	else if (rand_value <= diff+transp && transp)			//transparent
	{
		if(average <= 1.0f)
			*transparent = average;
		else
			*transparent = 1.0f;
		return(miPHOTON_TRANSPARENT);
	}
	else if (rand_value <= diff+transp+refl && refl)		//specular reflect
	{
		reflect->r = reflect->g = reflect->b = average;
		return(miPHOTON_REFLECT_SPECULAR);
	}
	else if (rand_value <= average && refr)					//specular transpmit
	{
		refract->r = refract->g = refract->b = average;
		return(miPHOTON_TRANSMIT_SPECULAR);
	}
	else
		return(miPHOTON_ABSORB);
}

// Absorbion coeff. from a color clipped to (0, 1]
//#define ABSORB(c) ((c)<=0.0f ? miHUGE_SCALAR : (c)>=1.0f ? 0.0f : -log(c))

/* Absorbion coeff. from a color clipped to (0, 1] */
inline void absorb(miColor* c, const miColor* a, miScalar dist)
{
   c->r *= (a->r <= 0.0f ? 0.0f : a->r >= 1.0f ? 1.0f : pow(a->r, dist));
   c->g *= (a->g <= 0.0f ? 0.0f : a->g >= 1.0f ? 1.0f : pow(a->g, dist));
   c->b *= (a->b <= 0.0f ? 0.0f : a->b >= 1.0f ? 1.0f : pow(a->b, dist));
}

inline void p_srand_color(miColor* rand_value, miUint index)
{
	unsigned short seed[3];
	while(index + 11 > USHRT_MAX)
		index -= USHRT_MAX;

	seed[0] = (unsigned short)index;
	seed[1] = (unsigned short)(index + 1);
	seed[2] = (unsigned short)(index + 2);
	rand_value->r = mi_erandom(seed);
	seed[0] = (unsigned short)(index + 3);
	seed[1] = (unsigned short)(index + 4);
	seed[2] = (unsigned short)(index + 5);
	rand_value->g = mi_erandom(seed);
	seed[0] = (unsigned short)(index + 6);
	seed[1] = (unsigned short)(index + 7);
	seed[2] = (unsigned short)(index + 8);
	rand_value->b = mi_erandom(seed);
	seed[0] = (unsigned short)(index + 9);
	seed[1] = (unsigned short)(index + 10);
	seed[2] = (unsigned short)(index + 11);
	rand_value->a = mi_erandom(seed);
}

inline void p_srand_float(float* rand_value, miUint index)
{
	unsigned short seed[3];
	while(index + 2 > USHRT_MAX)
		index -= USHRT_MAX;

	seed[0] = (unsigned short)index + 2;
	seed[1] = (unsigned short)(index);
	seed[2] = (unsigned short)(index + 1);
	*rand_value = mi_erandom(seed);
}

#endif  // p_utils_h
